"""
Class definitions for Relevance and related classes.
"""
from contextlib import contextmanager
from collections import defaultdict

import numpy as np

from openmdao.utils.general_utils import all_ancestors, _contains_all, get_rev_conns, env_truthy
from openmdao.utils.graph_utils import get_sccs_topo
from openmdao.utils.array_utils import array_hash
from openmdao.utils.om_warnings import issue_warning

_no_relevance = env_truthy('OPENMDAO_NO_RELEVANCE')


def get_relevance(model, of, wrt):
    """
    Return a Relevance object for the given design vars, and responses.

    Parameters
    ----------
    model : <Group>
        The top level group in the system hierarchy.
    of : dict
        Dictionary of 'of' variables.  Keys don't matter.
    wrt : dict
        Dictionary of 'wrt' variables.  Keys don't matter.

    Returns
    -------
    Relevance
        Relevance object.
    """
    if not model._use_derivatives or (not of and not wrt):
        # in this case, a permanently inactive relevance object is returned
        # (so the contents of 'of' and 'wrt' don't matter). Make them empty to avoid
        # unnecessary setup.
        of = {}
        wrt = {}

    key = (id(model), tuple(sorted(wrt)), tuple(sorted(of)))
    cache = model._problem_meta['relevance_cache']
    if key in cache:
        return cache[key]

    relevance = Relevance(model, wrt, of, model._problem_meta['rel_array_cache'])
    cache[key] = relevance
    return relevance


class Relevance(object):
    """
    Class that computes relevance based on a data flow graph.

    It determines current relevance based on the current set of forward and reverse seed variables.
    Initial relevance is determined by starting at a given seed and traversing the data flow graph
    in the specified direction to find all relevant variables and systems.  That information is
    then represented as a boolean array where True means the variable or system is relevant to the
    seed.  Relevance with respect to groups of seeds, for example, one forward seed vs. all reverse
    seeds, is determined by combining the boolean relevance arrays for the individual seeds in the
    following manner:  (fwd_array1 | fwd_array2 | ...) & (rev_array1 | rev_array2 | ...). In other
    words, the union of the fwd arrays is intersected with the union of the rev arrays.

    The full set of fwd and rev seeds must be set at initialization time.  At any point after that,
    the set of active seeds can be changed using the set_seeds method, but those seeds must be
    subsets of the full set of seeds.

    Parameters
    ----------
    model : <Group>
        The top level group in the system hierarchy.
    fwd_meta : dict
        Dictionary of design variable metadata.  Keys don't matter.
    rev_meta : dict
        Dictionary of response variable metadata.  Keys don't matter.
    rel_array_cache : dict
        Cache of relevance arrays stored by array hash.

    Attributes
    ----------
    _graph : <nx.DirectedGraph>
        Dependency graph.  Dataflow graph containing both variables and systems.
    _var2idx : dict
        dict of all variables in the graph mapped to the row index into the variable
        relevance array.
    _sys2idx : dict
        dict of all systems in the graph mapped to the row index into the system
        relevance array.
    _seed_vars : dict
        Maps direction to currently active seed variable names.
    _all_seed_vars : dict
        Maps direction to all seed variable names.
    _active : bool or None
        If True, relevance is active.  If False, relevance is inactive.  If None, relevance is
        uninitialized.
    _seed_var_map : dict
        Nested dict of the form {fwdseed(s): {revseed(s): var_array, ...}}.
        Keys that contain multiple seeds are sorted tuples of seed names.
    _seed_sys_map : dict
        Nested dict of the form {fwdseed(s): {revseed(s): sys_array, ...}}.
        Keys that contain multiple seeds are sorted tuples of seed names.
    _single_seed2relvars : dict
        Dict of the form {'fwd': {seed: var_array}, 'rev': ...} where each seed is a
        key and var_array is the variable relevance array for the given seed.
    _single_seed2relsys : dict
        Dict of the form {'fwd': {seed: sys_array}, 'rev': ...} where each seed is a
        key and var_array is the system relevance array for the given seed.
    _nonlinear_sets : dict
        Dict of the form {'pre': pre_rel_array, 'iter': iter_rel_array, 'post': post_rel_array}.
    _current_rel_varray : ndarray
        Array representing the variable relevance for the currently active seeds.
    _current_rel_sarray : ndarray
        Array representing the system relevance for the currently active seeds.
    _rel_array_cache : dict
        Cache of relevance arrays stored by array hash.
    _no_dv_responses : list
        List of responses that have no relevant design variables.
    _redundant_adjoint_systems : set or None
        Set of systems that may benefit from caching RHS arrays and solutions to avoid some linear
        solves.
    _seed_cache : dict
        Maps seed variable names to the source of the seed.
    _rel_array_cache : dict
        Cache of relevance arrays stored by array hash.
    empty : bool
        If True, relevance is empty and no relevance checking will be performed.
    """

    def __init__(self, model, fwd_meta, rev_meta, rel_array_cache):
        """
        Initialize all attributes.
        """
        assert model.pathname == '', "Relevance can only be initialized on the top level Group."

        # permanently disable relevance if _no_relevance is True
        self._active = False if _no_relevance else None
        self._rel_array_cache = rel_array_cache
        self._graph = model._dataflow_graph
        self._rel_array_cache = {}
        self._no_dv_responses = []
        self._redundant_adjoint_systems = None
        self._seed_cache = {}
        self.empty = False

        # seed var(s) for the current derivative operation
        self._seed_vars = {'fwd': (), 'rev': ()}
        # all seed vars for the entire derivative computation
        self._all_seed_vars = {'fwd': (), 'rev': ()}

        self._set_all_seeds(model, fwd_meta, rev_meta)

        self._current_rel_varray = None
        self._current_rel_sarray = None

        self._setup_nonlinear_relevance(model, fwd_meta, rev_meta)

        # _pre_components and _post_components will be empty unless the user has set the
        # 'group_by_pre_opt_post' option to True in the Problem.
        if model._pre_components or model._post_components:
            self._setup_nonlinear_sets(model)
        else:
            self._nonlinear_sets = {}

        # setting _active to False here will permanantly disable relevance checking for this
        # relevance object.  The only way to *temporarily* disable relevance is to use the
        # active() context manager.
        if not (fwd_meta and rev_meta):
            self._active = False

    def __repr__(self):
        """
        Return a string representation of the Relevance.

        Returns
        -------
        str
            String representation of the Relevance.
        """
        return f"Relevance({self._seed_vars}, active={self._active})"

    def _to_seed(self, names):
        """
        Return the seed from the given iter of names.

        Cache the given names iter if it is hashable.

        Parameters
        ----------
        names : iter of str
            Iterator over names.

        Returns
        -------
        tuple
            Key tuple for the given names.
        """
        try:
            return self._seed_cache[names]
        except TypeError:  # names is not hashable
            issue_warning("Relevance seeds should be hashable, but the following seed is not: "
                          f"{names}. It will be converted to a hashable form, but this could "
                          "cause performance issues.", category=RuntimeWarning)
            hashable = False
        except KeyError:  # names is not in the cache
            hashable = names

        try:
            names = [self._seed_cache[n] for n in names]
        except KeyError:
            raise KeyError(f"One or more of the relevance seeds '{names}' is invalid.")

        seeds = tuple(sorted(names))

        if hashable:
            self._seed_cache[hashable] = seeds

        return seeds

    def _get_cached_array(self, arr):
        """
        Return the cached array if it exists, otherwise return the input array after caching it.

        Parameters
        ----------
        arr : ndarray
            Array to be cached.

        Returns
        -------
        ndarray
            Cached array if it exists, otherwise the input array.
        """
        hash = array_hash(arr)
        if hash in self._rel_array_cache:
            return self._rel_array_cache[hash]
        else:
            self._rel_array_cache[hash] = arr

        return arr

    def _setup_nonlinear_sets(self, model):
        """
        Set up the nonlinear sets for relevance checking.

        Parameters
        ----------
        model : <Group>
            The top level group in the system hierarchy.
        """
        pre_systems = set()
        for compname in model._pre_components:
            pre_systems.update(all_ancestors(compname))
        if pre_systems:
            pre_systems.add('')  # include top level group

        post_systems = set()
        for compname in model._post_components:
            post_systems.update(all_ancestors(compname))
        if post_systems:
            post_systems.add('')

        pre_array = self._sys2rel_array(pre_systems)
        post_array = self._sys2rel_array(post_systems)

        if model._iterated_components is _contains_all:
            iter_array = np.ones(len(self._all_systems), dtype=bool)
        else:
            iter_systems = set()
            for compname in model._iterated_components:
                iter_systems.update(all_ancestors(compname))
            if iter_systems:
                iter_systems.add('')

            iter_array = self._sys2rel_array(iter_systems)

        self._nonlinear_sets = {'pre': pre_array, 'iter': iter_array, 'post': post_array}

    def _single_seed_array_iter(self, group, seed_meta, direction, all_systems):
        """
        Yield the relevance arrays for each individual seed and direction for variables and systems.

        The relevance arrays are boolean ndarrays of length nvars and nsystems, respectively.
        All of the variables and systems in the graph map to an index into these arrays and
        if the value at that index is True, then the variable or system is relevant to the seed.

        Parameters
        ----------
        group : <Group>
            The top level group in the system hierarchy.
        seed_meta : dict
            Dictionary of metadata for the seeds.
        direction : str
            Direction of the search for relevant variables.  'fwd' or 'rev'.
        all_systems : set
            Set of all systems in the graph.

        Yields
        ------
        str
            Name of the seed variable.
        bool
            True if the seed uses parallel derivative coloring.
        ndarray
            Boolean relevance array for the variables.
        ndarray
            Boolean relevance array for the systems.
        """
        nprocs = group.comm.size

        for meta in seed_meta.values():
            src = meta['source']
            local = nprocs > 1 and meta['parallel_deriv_color'] is not None
            if local:
                if src in group._var_abs2meta['output']:  # src is local
                    depnodes = self._dependent_nodes(src, direction, local=local)
                    group.comm.bcast(depnodes, root=group._owning_rank[src])
                else:
                    depnodes = group.comm.bcast(None, root=group._owning_rank[src])
            else:
                depnodes = self._dependent_nodes(src, direction, local=local)

            rel_systems = _vars2systems(depnodes)
            rel_vars = depnodes - all_systems

            yield (src, local, self._vars2rel_array(rel_vars), self._sys2rel_array(rel_systems))

    def _vars2rel_array(self, vars):
        """
        Return a relevance array for the given variables.

        Parameters
        ----------
        vars : iter of str
            Iterator over variable names.

        Returns
        -------
        ndarray
            Boolean relevance array.  True means name is relevant.
        """
        return self._names2rel_array(vars, self._var2idx)

    def _sys2rel_array(self, systems):
        """
        Return a relevance array for the given systems.

        Parameters
        ----------
        systems : iter of str
            Iterator over system names.

        Returns
        -------
        ndarray
            Boolean relevance array.  True means name is relevant.
        """
        return self._names2rel_array(systems, self._sys2idx)

    def _names2rel_array(self, names, names2inds):
        """
        Return a relevance array for the given names.

        Parameters
        ----------
        names : iter of str
            Iterator over names.
        names2inds : dict
            Dict of the form {name: index} where index is the index into the relevance array.

        Returns
        -------
        ndarray
            Boolean relevance array.  True means name is relevant.
        """
        rel_array = np.zeros(len(names2inds), dtype=bool)
        rel_array[[names2inds[n] for n in names]] = True

        return self._get_cached_array(rel_array)

    def _combine_relevance(self, fmap, fwd_seeds, rmap, rev_seeds):
        """
        Return the combined relevance arrays for the given seeds.

        Parameters
        ----------
        fmap : dict
            Dict of the form {seed: array} where array is the
            relevance arrays for the given seed.
        fwd_seeds : iter of str
            Iterator over forward seed variable names.
        rmap : dict
            Dict of the form {seed: array} where array is the
            relevance arrays for the given seed.
        rev_seeds : iter of str
            Iterator over reverse seed variable names.

        Returns
        -------
        ndarray
            Array representing the combined relevance arrays for the given seeds.
            The arrays are combined by taking the intersection of the relevance arrays for
            each fwd_seed/rev_seed pair and taking the union of each of those results.
        """
        combined = None
        for fseed in fwd_seeds:
            farr = fmap[fseed]
            for rseed in rev_seeds:
                if combined is None:
                    combined = farr & rmap[rseed]
                else:
                    combined |= (farr & rmap[rseed])

        return np.zeros(0, dtype=bool) if combined is None else self._get_cached_array(combined)

    def rel_vars_iter(self, rel_array, relevant=True):
        """
        Return an iterator of relevant variable names.

        Parameters
        ----------
        rel_array : ndarray
            Boolean relevance array.  True means name is relevant.
        relevant : bool
            If True, return only relevant names.  If False, return only irrelevant names.

        Yields
        ------
        str
            Name of the relevant variable.
        """
        yield from self._rel_names_iter(rel_array, self._var2idx, relevant)

    def _rel_names_iter(self, rel_array, all_names, relevant=True):
        """
        Return an iterator of names from the given relevance array.

        Parameters
        ----------
        rel_array : ndarray
            Boolean relevance array.  True means name is relevant.
        all_names : iter of str
            Iterator over the full set of names from the graph, either variables or systems.
        relevant : bool
            If True, return only relevant names.  If False, return only irrelevant names.

        Yields
        ------
        str
            Name from the given relevance array.
        """
        for n, rel in zip(all_names, rel_array):
            if rel == relevant:
                yield n

    def _set_all_seeds(self, group, fwd_meta, rev_meta):
        """
        Set the full list of seeds to be used to determine relevance.

        This should only be called once, at __init__ time.

        Parameters
        ----------
        group : <Group>
            The top level group in the system hierarchy.
        fwd_meta : dict
            Dictionary of metadata for forward derivatives.
        rev_meta : dict
            Dictionary of metadata for reverse derivatives.
        """
        fwd_seeds = []
        rev_seeds = []
        for name, meta in fwd_meta.items():
            src = meta['source']
            self._seed_cache[name] = src
            self._seed_cache[src] = src
            fwd_seeds.append(src)

        for name, meta in rev_meta.items():
            src = meta['source']
            self._seed_cache[name] = src
            self._seed_cache[src] = src
            rev_seeds.append(src)

        fwd_seeds = self._to_seed(tuple(fwd_seeds))
        rev_seeds = self._to_seed(tuple(rev_seeds))

        self._seed_var_map = seed_var_map = {}
        self._seed_sys_map = seed_sys_map = {}

        self._current_var_array = np.zeros(0, dtype=bool)
        self._current_sys_array = np.zeros(0, dtype=bool)

        self._all_seed_vars['fwd'] = fwd_seeds
        self._all_seed_vars['rev'] = rev_seeds

        self._single_seed2relvars = {'fwd': {}, 'rev': {}}
        self._single_seed2relsys = {'fwd': {}, 'rev': {}}

        if not fwd_meta or not rev_meta:
            self.empty = True
            self._sys2idx = {}
            self._var2idx = {}
            return

        # this set contains all variables and some or all components
        # in the graph.  Components are included if all of their outputs
        # depend on all of their inputs.
        all_vars = set()
        all_systems = {''}
        for node, data in self._graph.nodes(data=True):
            if 'type_' in data:
                all_vars.add(node)
                sysname = node.rpartition('.')[0]
                if sysname not in all_systems:
                    all_systems.update(all_ancestors(sysname))
            elif node not in all_systems:
                all_systems.update(all_ancestors(node))

        self._all_systems = all_systems
        all_vars = sorted(all_vars)

        # create mappings of var and system names to indices into the var/system
        # relevance arrays.
        self._sys2idx = {n: i for i, n in enumerate(sorted(all_systems))}
        self._var2idx = {n: i for i, n in enumerate(sorted(all_vars))}

        meta = {'fwd': fwd_meta, 'rev': rev_meta}

        # map each seed to its variable and system relevance arrays
        has_par_derivs = {}
        for io in ('fwd', 'rev'):
            for seed, local, var_array, sys_array in self._single_seed_array_iter(group, meta[io],
                                                                                  io, all_systems):
                self._single_seed2relvars[io][seed] = self._get_cached_array(var_array)
                self._single_seed2relsys[io][seed] = self._get_cached_array(sys_array)
                if local:
                    has_par_derivs[seed] = io

        # in seed_map, add keys for both fseed and (fseed,) and similarly for rseed
        # because both forms of keys may be used depending on the context.
        for fseed, fvarr in self._single_seed2relvars['fwd'].items():
            fsarr = self._single_seed2relsys['fwd'][fseed]
            seed_var_map[fseed] = seed_var_map[(fseed,)] = vsub = {}
            seed_sys_map[fseed] = seed_sys_map[(fseed,)] = ssub = {}
            for rseed, rvarr in self._single_seed2relvars['rev'].items():
                rsysarr = self._single_seed2relsys['rev'][rseed]
                vsub[rseed] = vsub[(rseed,)] = self._get_cached_array(fvarr & rvarr)
                ssub[rseed] = ssub[(rseed,)] = self._get_cached_array(fsarr & rsysarr)

        # now add entries for each (fseed, all_rseeds) and each (all_fseeds, rseed)
        for fsrc, farr in self._single_seed2relvars['fwd'].items():
            seed_var_map[fsrc][rev_seeds] = \
                self._combine_relevance(self._single_seed2relvars['fwd'], [fsrc],
                                        self._single_seed2relvars['rev'], rev_seeds)
            seed_sys_map[fsrc][rev_seeds] = \
                self._combine_relevance(self._single_seed2relsys['fwd'], [fsrc],
                                        self._single_seed2relsys['rev'], rev_seeds)

        seed_var_map[fwd_seeds] = {}
        seed_sys_map[fwd_seeds] = {}
        for rsrc, rarr in self._single_seed2relvars['rev'].items():
            seed_var_map[fwd_seeds][rsrc] = \
                self._combine_relevance(self._single_seed2relvars['fwd'], fwd_seeds,
                                        self._single_seed2relvars['rev'], [rsrc])
            seed_sys_map[fwd_seeds][rsrc] = \
                self._combine_relevance(self._single_seed2relsys['fwd'], fwd_seeds,
                                        self._single_seed2relsys['rev'], [rsrc])

        # now add 'full' relevance for all seeds
        seed_var_map[fwd_seeds][rev_seeds] = \
            self._combine_relevance(self._single_seed2relvars['fwd'], fwd_seeds,
                                    self._single_seed2relvars['rev'], rev_seeds)
        seed_sys_map[fwd_seeds][rev_seeds] = \
            self._combine_relevance(self._single_seed2relsys['fwd'], fwd_seeds,
                                    self._single_seed2relsys['rev'], rev_seeds)

        self._set_seeds(fwd_seeds, rev_seeds)

        if has_par_derivs:
            self._par_deriv_err_check(group, rev_meta, fwd_meta)

        found = set()
        for fsrc, farr in self._single_seed2relvars['fwd'].items():
            for rsrc, rarr in self._single_seed2relvars['rev'].items():
                if rsrc not in found:
                    if (farr & rarr)[self._var2idx[fsrc]]:
                        found.add(rsrc)

        self._no_dv_responses = \
            [rsrc for rsrc in self._single_seed2relvars['rev'] if rsrc not in found]

    def get_redundant_adjoint_systems(self):
        """
        Find any systems that depend on responses that depend on other responses.

        If any are found, it may be worthwhile to cache RHS arrays and solutions in order to avoid
        some linear solves.

        Returns
        -------
        dict
            Mapping of systems to the set of adjoints that can cause unnecessary linear solves.
        """
        if self._redundant_adjoint_systems is None:
            self._redundant_adjoint_systems = defaultdict(set)
            resp2resp_deps = set()
            for rsrc, arr1 in self._single_seed2relvars['rev'].items():
                for rsrc2 in self._single_seed2relvars['rev']:
                    if rsrc2 != rsrc:
                        if arr1[self._var2idx[rsrc2]]:
                            # add dependent pairs of responses
                            resp2resp_deps.add((rsrc, rsrc2))

            if resp2resp_deps:
                fsystems = self._seed_sys_map[self._all_seed_vars['fwd']]
                for rsrc, rsrc2 in resp2resp_deps:
                    relarr = fsystems[rsrc] & fsystems[rsrc2]  # intersection
                    for relevant_system in self._rel_names_iter(relarr, self._sys2idx):
                        self._redundant_adjoint_systems[relevant_system].update((rsrc, rsrc2))

        return self._redundant_adjoint_systems

    @contextmanager
    def active(self, active):
        """
        Context manager for temporarily deactivating relevance.

        Note that if this relevance object is already inactive, this context manager will have no
        effect, i.e., calling this with active=True will not activate an inactive relevance object,
        but calling it with active=False will deactivate an active relevance object.

        The only way to activate an otherwise inactive relevance object is to use the
        all_seeds_active, seeds_active, or nonlinear_active context managers and this will only
        work if _active is None or True.

        Parameters
        ----------
        active : bool
            If True, activate relevance.  If False, deactivate relevance.

        Yields
        ------
        None
        """
        if not self._active:  # if already inactive from higher level, don't change it
            yield
        else:
            save = self._active
            self._active = active
            try:
                yield
            finally:
                self._active = save

    def relevant_vars(self, name, direction, inputs=True, outputs=True):
        """
        Return a set of variables relevant to the given dv/response in the given direction.

        Parameters
        ----------
        name : str
            Name of the variable of interest.
        direction : str
            Direction of the search for relevant variables.  'fwd' or 'rev'.
        inputs : bool
            If True, include inputs.
        outputs : bool
            If True, include outputs.

        Returns
        -------
        set
            Set of the relevant variables.
        """
        names = self._rel_names_iter(self._single_seed2relvars[direction][name], self._var2idx)
        if inputs and outputs:
            return set(names)
        elif inputs:
            return self._apply_node_filter(names, _is_input)
        elif outputs:
            return self._apply_node_filter(names, _is_output)
        else:
            return set()

    @contextmanager
    def all_seeds_active(self):
        """
        Context manager where all seeds are active.

        If _active is False, this will have no effect.

        Yields
        ------
        None
        """
        # if already inactive from higher level, or 'active' parameter is False, don't change it
        if self._active is False:
            yield
        else:
            save = {'fwd': self._seed_vars['fwd'], 'rev': self._seed_vars['rev']}
            save_active = self._active
            self._active = True
            self._set_seeds(self._all_seed_vars['fwd'], self._all_seed_vars['rev'])
            try:
                yield
            finally:
                self._seed_vars = save
                self._active = save_active

    @contextmanager
    def seeds_active(self, fwd_seeds=None, rev_seeds=None):
        """
        Context manager where the specified seeds are active.

        If _active is False, this will have no effect.

        Parameters
        ----------
        fwd_seeds : iter of str or None
            Iterator over forward seed variable names. If None use current active seeds.
        rev_seeds : iter of str or None
            Iterator over reverse seed variable names. If None use current active seeds.

        Yields
        ------
        None
        """
        if self._active is False:  # if already inactive from higher level, don't change anything
            yield
        else:
            save = {'fwd': self._seed_vars['fwd'], 'rev': self._seed_vars['rev']}
            save_active = self._active
            self._active = True
            if fwd_seeds is None:
                fwd_seeds = self._seed_vars['fwd']
            if rev_seeds is None:
                rev_seeds = self._seed_vars['rev']
            self._set_seeds(fwd_seeds, rev_seeds)
            try:
                yield
            finally:
                self._seed_vars = save
                self._active = save_active

    @contextmanager
    def nonlinear_active(self, name, active=True):
        """
        Context manager for activating a subset of systems using 'pre', 'post', or 'iter'.

        Parameters
        ----------
        name : str
            Name of the set to activate.
        active : bool
            If False, relevance is temporarily deactivated.

        Yields
        ------
        None
        """
        if not active or self._active is False or name not in self._nonlinear_sets:
            yield
        else:
            save_active = self._active
            save_relarray = self._current_rel_sarray
            self._active = True
            self._current_rel_sarray = self._nonlinear_sets[name]

            try:
                yield
            finally:
                self._active = save_active
                self._current_rel_sarray = save_relarray

    def _set_seeds(self, fwd_seeds, rev_seeds):
        """
        Set the seed(s) to determine relevance for a given variable in a given direction.

        Parameters
        ----------
        fwd_seeds : frozenset
            Set of forward seed variable names.
        rev_seeds : frozenset
            Set of reverse seed variable names.
        """
        fwd_seeds = self._to_seed(fwd_seeds)
        rev_seeds = self._to_seed(rev_seeds)

        self._seed_vars['fwd'] = fwd_seeds
        self._seed_vars['rev'] = rev_seeds

        self._current_rel_varray = self._get_rel_array(self._seed_var_map,
                                                       self._single_seed2relvars,
                                                       fwd_seeds, rev_seeds)
        if self._current_rel_varray.size == 0:
            self._active = False

        self._current_rel_sarray = self._get_rel_array(self._seed_sys_map,
                                                       self._single_seed2relsys,
                                                       fwd_seeds, rev_seeds)

    def _get_rel_array(self, seed_map, single_seed2rel, fwd_seeds, rev_seeds):
        """
        Return the combined relevance array for the given seeds.

        If it doesn't exist, create it.

        Parameters
        ----------
        seed_map : dict
            Dict of the form {fwdseed: {revseed: rel_arrays}}.
        single_seed2rel : dict
            Dict of the form {'fwd': {seed: rel_array}, 'rev': ...} where each seed is a key and
            rel_array is the relevance array for the given seed.
        fwd_seeds : str or sorted tuple of str
            Iterator over forward seed variable names.
        rev_seeds : str or sorted tuple of str
            Iterator over reverse seed variable names.

        Returns
        -------
        ndarray
            Array representing the combined relevance arrays for the given seeds.
        """
        try:
            return seed_map[fwd_seeds][rev_seeds]
        except KeyError:
            # don't have a relevance array for this seed combo, so create it
            relarr = self._combine_relevance(single_seed2rel['fwd'], fwd_seeds,
                                             single_seed2rel['rev'], rev_seeds)
            if fwd_seeds not in seed_map:
                seed_map[fwd_seeds] = {}
            seed_map[fwd_seeds][rev_seeds] = relarr

        return relarr

    def is_relevant(self, name):
        """
        Return True if the given variable is relevant.

        Parameters
        ----------
        name : str
            Name of the variable.

        Returns
        -------
        bool
            True if the given variable is relevant.
        """
        if not self._active:
            return True

        return self._current_rel_varray[self._var2idx[name]]

    def any_relevant(self, names):
        """
        Return True if any of the given variables are relevant.

        Parameters
        ----------
        names : iter of str
            Iterator over variable names.

        Returns
        -------
        bool
            True if any of the given variables are relevant.
        """
        if self.empty:
            return False
        if not self._active:
            return True

        for n in names:
            if self._current_rel_varray[self._var2idx[n]]:
                return True
        return False

    def is_relevant_system(self, name):
        """
        Return True if the given named system is relevant.

        Returns False if system has no subsystems with outputs.

        Parameters
        ----------
        name : str
            Name of the System.

        Returns
        -------
        bool
            True if the given system is relevant.
        """
        if not self._active:
            return True

        try:
            return self._current_rel_sarray[self._sys2idx[name]]
        except KeyError:
            return False

    def filter(self, systems, relevant=True):
        """
        Filter the given iterator of systems to only include those that are relevant.

        Parameters
        ----------
        systems : iter of Systems
            Iterator over systems.
        relevant : bool
            If True, return only relevant systems.  If False, return only irrelevant systems.

        Yields
        ------
        System
            Relevant system.
        """
        if self._active:
            for system in systems:
                if relevant == self.is_relevant_system(system.pathname):
                    yield system
        elif relevant:
            yield from systems

    def iter_seed_pair_relevance(self, fwd_seeds=None, rev_seeds=None, inputs=False, outputs=False):
        """
        Yield all relevant variables for each pair of seeds.

        Parameters
        ----------
        fwd_seeds : iter of str or None
            Iterator over forward seed variable names. If None use current registered seeds.
        rev_seeds : iter of str or None
            Iterator over reverse seed variable names. If None use current registered seeds.
        inputs : bool
            If True, include inputs.
        outputs : bool
            If True, include outputs.

        Yields
        ------
        set
            Set of names of relevant variables.
        """
        filt = _get_io_filter(inputs, outputs)
        if filt is True:  # everything is filtered out
            return

        if fwd_seeds is None:
            fwd_seeds = self._seed_vars['fwd']
        if rev_seeds is None:
            rev_seeds = self._seed_vars['rev']

        if isinstance(fwd_seeds, str):
            fwd_seeds = [fwd_seeds]
        if isinstance(rev_seeds, str):
            rev_seeds = [rev_seeds]

        for seed in fwd_seeds:
            for rseed in rev_seeds:
                inter = self._get_rel_array(self._seed_var_map, self._single_seed2relvars,
                                            seed, rseed)
                if np.any(inter):
                    inter = self._rel_names_iter(inter, self._var2idx)
                    yield seed, rseed, self._apply_node_filter(inter, filt)

    def _apply_node_filter(self, names, filt):
        """
        Return only the nodes from the given set of nodes that pass the given filter.

        Parameters
        ----------
        names : iter of str
            Iterator of node names.
        filt : callable
            Filter function taking a graph node as an argument and returning True if the node
            should be included in the output.  If True, no filtering is done.  If False, the
            returned set will be empty.

        Returns
        -------
        set
            Set of node names that passed the filter.
        """
        if not filt:  # no filtering needed
            if isinstance(names, set):
                return names
            return set(names)
        elif filt is True:
            return set()

        # filt is a function.  Apply it to named graph nodes.
        return set(self._filter_nodes_iter(names, filt))

    def _filter_nodes_iter(self, names, filt):
        """
        Return only the nodes from the given set of nodes that pass the given filter.

        Parameters
        ----------
        names : iter of str
            Iterator over node names.
        filt : callable
            Filter function taking a graph node as an argument and returning True if the node
            should be included in the output.

        Yields
        ------
        str
            Node name that passed the filter.
        """
        nodes = self._graph.nodes
        for n in names:
            if filt(nodes[n]):
                yield n

    def _all_relevant(self, fwd_seeds, rev_seeds, inputs=True, outputs=True):
        """
        Return all relevant inputs, outputs, and systems for the given seeds.

        This is primarily used as a convenience function for testing and is not particularly
        efficient.

        Parameters
        ----------
        fwd_seeds : iter of str
            Iterator over forward seed variable names.
        rev_seeds : iter of str
            Iterator over reverse seed variable names.
        inputs : bool
            If True, include inputs.
        outputs : bool
            If True, include outputs.

        Returns
        -------
        tuple
            (set of relevant inputs, set of relevant outputs, set of relevant systems)
            If a given inputs/outputs is False, the corresponding set will be empty. The
            returned systems will be the set of all systems containing any
            relevant variables based on the values of inputs and outputs, i.e. if outputs is False,
            the returned systems will be the set of all systems containing any relevant inputs.
        """
        relevant_vars = set()
        for _, _, relvars in self.iter_seed_pair_relevance(fwd_seeds, rev_seeds, inputs, outputs):
            relevant_vars.update(relvars)
        relevant_systems = _vars2systems(relevant_vars)

        inputs = set(self._filter_nodes_iter(relevant_vars, _is_input))
        outputs = set(self._filter_nodes_iter(relevant_vars, _is_output))

        return inputs, outputs, relevant_systems

    def _dependent_nodes(self, start, direction, local=False):
        """
        Return set of all connected nodes in the given direction starting at the given node.

        Parameters
        ----------
        start : str
            Name of the starting node.
        direction : str
            If 'fwd', traverse downstream.  If 'rev', traverse upstream.
        local : bool
            If True, include only local variables.

        Returns
        -------
        set
            Set of all dependent nodes.
        """
        if start in self._graph:
            if local and not self._graph.nodes[start]['local']:
                return set()

            if direction == 'fwd':
                fnext = self._graph.successors
            elif direction == 'rev':
                fnext = self._graph.predecessors
            else:
                raise ValueError("direction must be 'fwd' or 'rev'")

            stack = [start]
            visited = {start}

            while stack:
                src = stack.pop()
                for tgt in fnext(src):
                    if tgt not in visited:
                        if local:
                            node = self._graph.nodes[tgt]
                            # stop local traversal at the first non-local node
                            if 'local' in node and not node['local']:
                                return visited

                        visited.add(tgt)
                        stack.append(tgt)

            return visited

        return set()

    def _par_deriv_err_check(self, group, responses, desvars):
        pd_err_chk = defaultdict(dict)
        mode = group._problem_meta['mode']  # 'fwd', 'rev', or 'auto'

        if mode in ('fwd', 'auto'):
            for desvar, response, relset in self.iter_seed_pair_relevance(inputs=True):
                if desvar in desvars and self._graph.nodes[desvar]['local']:
                    dvcolor = desvars[desvar]['parallel_deriv_color']
                    if dvcolor:
                        pd_err_chk[dvcolor][desvar] = relset

        if mode in ('rev', 'auto'):
            for desvar, response, relset in self.iter_seed_pair_relevance(outputs=True):
                if response in responses and self._graph.nodes[response]['local']:
                    rescolor = responses[response]['parallel_deriv_color']
                    if rescolor:
                        pd_err_chk[rescolor][response] = relset

        # check to make sure we don't have any overlapping dependencies between vars of the
        # same color
        errs = {}
        for pdcolor, dct in pd_err_chk.items():
            for vname, relset in dct.items():
                for n, nds in dct.items():
                    if vname != n and relset.intersection(nds):
                        if pdcolor not in errs:
                            errs[pdcolor] = []
                        errs[pdcolor].append(vname)

        all_errs = group.comm.allgather(errs)
        msg = []
        for errdct in all_errs:
            for color, names in errdct.items():
                vtype = 'design variable' if mode == 'fwd' else 'response'
                msg.append(f"Parallel derivative color '{color}' has {vtype}s "
                           f"{sorted(names)} with overlapping dependencies on the same rank.")

        if msg:
            raise RuntimeError('\n'.join(msg))

    def _setup_nonlinear_relevance(self, model, designvars, responses):
        """
        Set up the iteration lists containing the pre, iterated, and post subsets of systems.

        This should only be called on the top level Group.

        Parameters
        ----------
        model : <Group>
            The top level group in the system hierarchy.
        designvars : dict
            A dict of all design variables from the model.
        responses : dict
            A dict of all responses from the model.
        """
        # don't redo this if it's already done
        if model._pre_components is not None:
            return

        if not designvars or not responses or not model._problem_meta['group_by_pre_opt_post']:
            return

        model._pre_components = set()
        model._post_components = set()
        model._iterated_components = _contains_all

        # keep track of Groups with nonlinear solvers that use gradients (like Newton) and certain
        # linear solvers like DirectSolver. These groups and all systems they contain must be
        # grouped together into the same iteration list.
        grad_groups = set()
        always_opt = set()
        model._get_relevance_modifiers(grad_groups, always_opt)

        if '' in grad_groups:
            issue_warning("The top level group has a nonlinear solver that computes gradients, so "
                          "the entire model will be included in the optimization iteration.")
            return

        dvs = [meta['source'] for meta in designvars.values()]
        responses = [meta['source'] for meta in responses.values()]
        responses = set(responses)  # get rid of dups due to aliases

        graph = model.compute_sys_graph(comps_only=True, add_edge_info=False)

        auto_dvs = [dv for dv in dvs if dv.startswith('_auto_ivc.')]
        dv0 = auto_dvs[0] if auto_dvs else dvs[0].rpartition('.')[0]

        if auto_dvs:
            rev_conns = get_rev_conns(model._conn_global_abs_in2out)

            # add nodes for any auto_ivc vars that are dvs and connect to downstream component(s)
            for dv in auto_dvs:
                graph.add_node(dv, type_='output')
                inps = rev_conns.get(dv, ())
                for inp in inps:
                    inpcomp = inp.rpartition('.')[0]
                    graph.add_edge(dv, inpcomp)

        # One way to determine the contents of the pre/opt/post sets is to add edges from the
        # response variables to the design variables and vice versa, then find the strongly
        # connected components of the resulting graph.  get_sccs_topo returns the strongly
        # connected components in topological order, so we can use it to give us pre, iterated,
        # and post subsets of the systems.

        # add edges between response comps and design vars/comps to form a strongly
        # connected component for all nodes involved in the optimization iteration.
        for res in responses:
            resnode = res.rpartition('.')[0]
            for dv in dvs:
                dvnode = dv.rpartition('.')[0]
                if dvnode == '_auto_ivc':
                    # var node exists in graph so connect it to resnode
                    dvnode = dv  # use var name not comp name

                graph.add_edge(resnode, dvnode)
                graph.add_edge(dvnode, resnode)

        # loop 'always_opt' components into all responses to force them to be relevant during
        # optimization.
        for opt_sys in always_opt:
            for response in responses:
                rescomp = response.rpartition('.')[0]
                graph.add_edge(opt_sys, rescomp)
                graph.add_edge(rescomp, opt_sys)

        groups_added = set()

        if grad_groups:
            remaining = set(grad_groups)
            for name in sorted(grad_groups, key=lambda x: x.count('.')):
                prefix = name + '.'
                match = {n for n in remaining if n.startswith(prefix)}
                remaining -= match

            gradlist = '\n'.join(sorted(remaining))
            issue_warning("The following groups have a nonlinear solver that computes gradients "
                          f"and will be treated as atomic for the purposes of determining "
                          f"which systems are included in the optimization iteration: "
                          f"\n{gradlist}\n")

            # remaining groups are not contained within a higher level nl solver
            # using gradient group, so make new connections to/from them to
            # all systems that they contain.  This will force them to be
            # treated as 'atomic' within the graph, so that if they contain
            # any dv or response systems, or if their children are connected to
            # both dv *and* response systems, then all systems within them will
            # be included in the 'opt' set.  Note that this step adds some group nodes
            # to the graph where before it only contained component nodes and auto_ivc
            # var nodes.
            edges_to_add = []
            for grp in remaining:
                prefix = grp + '.'
                for node in graph:
                    if node.startswith(prefix):
                        groups_added.add(grp)
                        edges_to_add.append((grp, node))
                        edges_to_add.append((node, grp))

            graph.add_edges_from(edges_to_add)

        # this gives us the strongly connected components in topological order
        sccs = get_sccs_topo(graph)

        pre = addto = set()
        post = set()
        iterated = set()
        for strong_con in sccs:
            # because the sccs are in topological order and all design vars and
            # responses are in the iteration set, we know that until we
            # see a design var or response, we're in the pre-opt set.  Once we
            # see a design var or response, we're in the iterated set.  Once
            # we see an scc without a design var or response, we're in the
            # post-opt set.
            if dv0 in strong_con:
                for s in strong_con:
                    if 'type_' in graph.nodes[s]:
                        s = s.rpartition('.')[0]
                    if s not in iterated:
                        iterated.add(s)
                addto = post
            else:
                for s in strong_con:
                    if 'type_' in graph.nodes[s]:
                        s = s.rpartition('.')[0]
                    if s not in addto:
                        addto.add(s)

        auto_ivc = model._auto_ivc
        auto_dvs = set(auto_dvs)
        rev_conns = get_rev_conns(model._conn_global_abs_in2out)
        if '_auto_ivc' not in pre:
            in_pre = False
            for vname in auto_ivc._var_abs2prom['output']:
                if vname not in auto_dvs:
                    for tgt in rev_conns[vname]:
                        tgtcomp = tgt.rpartition('.')[0]
                        if tgtcomp in pre:
                            in_pre = True
                            break
                    if in_pre:
                        break
            if in_pre:
                pre.add('_auto_ivc')

        # if 'pre' contains nothing but _auto_ivc, then just make it empty
        if len(pre) == 1 and '_auto_ivc' in pre:
            pre.discard('_auto_ivc')

        model._pre_components = pre - groups_added
        model._post_components = post - groups_added
        model._iterated_components = iterated - groups_added

        # it's possible that some components could be in pre on some ranks and post in others
        # if they are not connected in any way to any components in the iterated set, so we
        # need to pick a rank and bcast the final pre and post sets to all ranks to ensure
        # consistency.
        if model.comm.size > 1:
            pre, post = model.comm.bcast((model._pre_components, model._post_components), root=0)
            model._pre_components = pre
            model._post_components = post

    def list_relevance(self, relevant=True, type='system'):
        """
        Return a list of relevant variables and systems for the given seeds.

        Parameters
        ----------
        relevant : bool
            If True, return only relevant variables and systems.  If False, return only irrelevant
            variables and systems.
        type : str
            If 'system', return only system names.  If 'var', return only variable names.

        Returns
        -------
        list of str
            List of (ir)relevant variables or systems.
        """
        if type == 'system':
            it = self._rel_names_iter(self._current_rel_sarray, self._sys2idx, relevant)
        else:
            it = self._rel_names_iter(self._current_rel_varray, self._var2idx, relevant)

        return list(it)


def _vars2systems(nameiter):
    """
    Return a set of all systems containing the given variables or components.

    This includes all ancestors of each system, including ''.

    Parameters
    ----------
    nameiter : iter of str
        Iterator of variable or component pathnames.

    Returns
    -------
    set
        Set of system pathnames.
    """
    systems = {''}  # root group is always there
    for name in nameiter:
        sysname = name.rpartition('.')[0]
        if sysname not in systems:
            systems.update(all_ancestors(sysname))

    return systems


def _get_io_filter(inputs, outputs):
    if inputs and outputs:
        return False  # no filtering needed
    elif inputs:
        return _is_input
    elif outputs:
        return _is_output
    else:
        return True  # filter out everything


def _is_input(node):
    return node['type_'] == 'input'


def _is_output(node):
    return node['type_'] == 'output'


def _dump_seed_map(seed_map):
    """
    Print the contents of the given seed_map for debugging.

    Parameters
    ----------
    seed_map : dict
        Dict of the form {fwdseed: {revseed: rel_arrays}}.
    """
    for fseed, relmap in seed_map.items():
        for rseed, relarr in relmap.items():
            print(f'({fseed}, {rseed}) {np.asarray(relarr, dtype=np.uint8)}')



"""
An ExplicitComponent that uses JAX for derivatives.
"""

import sys
from openmdao.core.explicitcomponent import ExplicitComponent
from openmdao.utils.jax_utils import jax
from openmdao.utils.om_warnings import issue_warning


class JaxExplicitComponent(ExplicitComponent):
    """
    Base class for explicit components when using JAX for derivatives.

    Parameters
    ----------
    fallback_derivs_method : str
        The method to use if JAX is not available. Default is 'fd'.
    **kwargs : dict
        Additional arguments to be passed to the base class.
    """

    def __init__(self, fallback_derivs_method='fd', **kwargs):  # noqa
        if sys.version_info < (3, 9):
            raise RuntimeError("JaxExplicitComponent requires Python 3.9 or newer.")
        super().__init__(**kwargs)
        # if derivs_method is explicitly passed in, just use it
        if 'derivs_method' in kwargs:
            return

        if jax:
            self.options['derivs_method'] = 'jax'
        else:
            issue_warning(f"{self.msginfo}: JAX is not available, so '{fallback_derivs_method}' "
                          "will be used for derivatives.")
            self.options['derivs_method'] = fallback_derivs_method




import openmdao.api as om

# Note: The following class definitions are used in feature docs


class RectangleComp(om.ExplicitComponent):
    """
    A simple Explicit Component that computes the area of a rectangle.
    """

    def setup(self):
        self.add_input('length', val=1.)
        self.add_input('width', val=1.)
        self.add_output('area', val=1.)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        outputs['area'] = inputs['length'] * inputs['width']


class RectangleCompWithTags(om.ExplicitComponent):
    """
    A simple Explicit Component that also has input and output with tags.
    """

    def setup(self):
        self.add_input('length', val=1., tags=["tag1"])
        self.add_input('width', val=1., tags=["tag2"])
        self.add_output('area', val=1., tags=["tag1"])

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        outputs['area'] = inputs['length'] * inputs['width']


class RectanglePartial(RectangleComp):

    def compute_partials(self, inputs, partials):
        partials['area', 'length'] = inputs['width']
        partials['area', 'width'] = inputs['length']


class RectangleJacVec(RectangleComp):

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        if mode == 'fwd':
            if 'area' in d_outputs:
                if 'length' in d_inputs:
                    d_outputs['area'] += inputs['width'] * d_inputs['length']
                if 'width' in d_inputs:
                    d_outputs['area'] += inputs['length'] * d_inputs['width']
        elif mode == 'rev':
            if 'area' in d_outputs:
                if 'length' in d_inputs:
                    d_inputs['length'] += inputs['width'] * d_outputs['area']
                if 'width' in d_inputs:
                    d_inputs['width'] += inputs['length'] * d_outputs['area']


class RectangleGroup(om.Group):

    def setup(self):
        self.add_subsystem('comp1', RectanglePartial(), promotes_inputs=['width', 'length'])
        self.add_subsystem('comp2', RectangleJacVec(), promotes_inputs=['width', 'length'])



"""
3 Bar Truss Problem.

- 3 continuous design variables -  Cross sectional area of each bar.

- 3 discrete material choices for each bar. Complete enumeration requires 64 continuous optimization.

The global optima is 3,3,x (x can be anything), with mass = 5.28 kg
"""

import numpy as np

import openmdao.api as om


def stress_calc(A, E):
    """
    Calculate stress in a 3-bar truss given area and Young's modulus.

    Parameters
    ----------
    A : float or ndarray
        Area of the bar.
    E : float or ndarray
        Young's Modules of the bar.

    Returns
    -------
    float or ndarray
        Stress in the bar.
    """
    P  = 120000.0
    L = np.array([np.sqrt(1.2**2 + 1.2**2), 1.2, np.sqrt(1.2**2 + 1.2**2)])

    #Local stiffness matrix
    theta1 = -45.0*np.pi/180.0
    theta2 = -90.0*np.pi/180.0
    theta3 = -135*np.pi/180.0

    K0 = (E[0]*A[0]/L[0])*np.dot(np.array([[np.cos(theta1), np.sin(theta1)]]).T,
                                          np.array([[np.cos(theta1), np.sin(theta1)]]))
    K1 = (E[1]*A[1]/L[1])*np.dot(np.array([[np.cos(theta2), np.sin(theta2)]]).T,
                                          np.array([[np.cos(theta2), np.sin(theta2)]]))
    K2 = (E[2]*A[2]/L[2])*np.dot(np.array([[np.cos(theta3), np.sin(theta3)]]).T,
                                          np.array([[np.cos(theta3), np.sin(theta3)]]))

    # Global (total) stiffness matrix
    K = K0 + K1 + K2

    # Load vector
    theta4 = -65.0*np.pi/180.0
    p = P*np.array([[np.cos(theta4), np.sin(theta4)]]).T
    # Displacement matrix
    u = np.dot(np.linalg.inv(K), p)

    #Delta change in length
    DL = np.zeros([3])
    DL[0] = np.sqrt((-L[0]*np.cos(theta1) - u[0])**2 + (-L[0]*np.sin(theta1) - u[1])**2).item() - L[0]
    DL[1] = np.sqrt((-L[1]*np.cos(theta2) - u[0])**2 + (-L[1]*np.sin(theta2) - u[1])**2).item() - L[1]
    DL[2] = np.sqrt((-L[2]*np.cos(theta3) - u[0])**2 + (-L[2]*np.sin(theta3) - u[1])**2).item() - L[2]

    #Stress in each element
    sigma = E*DL/L
    return sigma


class ThreeBarTruss(om.ExplicitComponent):
    """
    3 Bar truss problem with 3 continuous design variables and 3 discrete
    material choices. Material chosen as follows:

            1 Aluminum
            2 Titanium
            3 Steel
            4 Nickel

    This component calculates the total mass of the truss.
    """

    def setup(self):
        # Continuous Inputs
        self.add_input('area1', 0.0, units='cm**2',
                       desc='Cross-sectional area of beam 1')
        self.add_input('area2', 0.0, units='cm**2',
                       desc='Cross-sectional area of beam 2')
        self.add_input('area3', 0.0, units='cm**2',
                       desc='Cross-sectional area of beam 3')

        # Discrete Inputs
        self.add_input('mat1', 1, #lower=1, upper=4,
                       desc='Material ID of beam 1')
        self.add_input('mat2', 1, #lower=1, upper=4,
                       desc='Material ID of beam 2')
        self.add_input('mat3', 1, #lower=1, upper=4,
                       desc='Material ID of beam 3')

        # Outputs
        self.add_output('mass', val=0.0)
        self.add_output('stress', val=np.zeros((3, )))

        self.rho = { 1 : 2700.0,
                     2 : 4500.0,
                     3 : 7872.0,
                     4 : 8800.0 }

        self.E = { 1 : 68.9e9,
                   2 : 116.0e9,
                   3 : 205.0e9,
                   4 : 207.0e9 }

        self.sigma_y = { 1 : 55.2e6,
                         2 : 140.0e6,
                         3 : 285.0e6,
                         4 : 59.0e6 }

    def setup_partials(self):
        self.declare_partials(of='*', wrt='area*', method='fd')

    def compute(self, inputs, outputs):
        """
        Define the function f(xI, xC).

        Here xI is integer and xC is continuous.
        """

        # Convert areas to m**2
        area1 = inputs['area1']*.0001
        area2 = inputs['area2']*.0001
        area3 = inputs['area3']*.0001
        mat1 = int(inputs['mat1'].item())
        mat2 = int(inputs['mat2'].item())
        mat3 = int(inputs['mat3'].item())
        area1 *= 2.

        len1 = np.sqrt(1.2**2 + 1.2**2)
        len2 = 1.2
        len3 = np.sqrt(1.2**2 + 1.2**2)

        rho1 = self.rho[mat1]
        rho2 = self.rho[mat2]
        rho3 = self.rho[mat3]

        outputs['mass'] = rho1*area1*len1 + rho2*area2*len2 + rho3*area3*len3

        E1 = self.E[mat1]
        E2 = self.E[mat2]
        E3 = self.E[mat3]
        sigma_y1 = self.sigma_y[mat1]
        sigma_y2 = self.sigma_y[mat2]
        sigma_y3 = self.sigma_y[mat3]

        E = np.array([E1, E2, E3])
        A = np.array([area1, area2, area3])
        sigma_y = np.array([sigma_y1, sigma_y2, sigma_y3])

        sigma = stress_calc(A, E)
        outputs['stress'] = (np.abs(sigma)/sigma_y)


class ThreeBarTrussVector(om.ExplicitComponent):
    """
    3 Bar truss problem with 3 continuous design variables and 3 discrete
    material choices. Material chosen as follows:

            1 Aluminum
            2 Titanium
            3 Steel
            4 Nickel

    This component calculates the total mass of the truss.

    This version places all areas in a single vector and all materials in
    a single vector for test purporses.
    """

    def setup(self):
        # Continuous Inputs
        self.add_input('area', np.array([0.0, 0.0, 0.0]), units='cm**2',
                       desc='Cross-sectional area of beams')

        # Discrete Inputs
        self.add_input('mat', np.array([1, 1, 1]), #lower=1, upper=4,
                       desc='Material ID of beams')

        # Outputs
        self.add_output('mass', val=0.0)
        self.add_output('stress', val=np.zeros((3, )))

        self.rho = { 1 : 2700.0,
                     2 : 4500.0,
                     3 : 7872.0,
                     4 : 8800.0 }

        self.E = { 1 : 68.9e9,
                   2 : 116.0e9,
                   3 : 205.0e9,
                   4 : 207.0e9 }

        self.sigma_y = { 1 : 55.2e6,
                         2 : 140.0e6,
                         3 : 285.0e6,
                         4 : 59.0e6 }

    def setup_partials(self):
        self.declare_partials(of='*', wrt='area', method='fd')

    def compute(self, inputs, outputs):
        """ Define the function f(xI, xC)
        Here xI is integer and xC is continuous"""

        # Area convert to m**2
        area = inputs['area']*.0001
        mat = inputs['mat']

        length = np.array([np.sqrt(1.2**2 + 1.2**2),
                           1.2,
                           np.sqrt(1.2**2 + 1.2**2)])

        rho = np.zeros((3, ))
        try:
            rho[0] = self.rho[mat[0]]
        except Exception:
            pass
        rho[1] = self.rho[mat[1]]
        rho[2] = self.rho[mat[2]]

        outputs['mass'] = np.sum(rho*area*length)

        E = np.zeros((3, ))
        E[0] = self.E[mat[0]]
        E[1] = self.E[mat[1]]
        E[2] = self.E[mat[2]]
        sigma_y = np.zeros((3, ))

        sigma_y[0] = self.sigma_y[mat[0]]
        sigma_y[1] = self.sigma_y[mat[1]]
        sigma_y[2] = self.sigma_y[mat[2]]

        sigma = stress_calc(area, E)
        outputs['stress'] = (np.abs(sigma)/sigma_y)



"""
Visualization of data functions.
"""
import numpy as np


def partial_deriv_plot(of, wrt, check_partials_data, title=None, jac_method='J_fwd', tol=1e-10,
                       binary=True):
    """
    Visually examine the computed and finite differenced Jacobians.

    Parameters
    ----------
    of : str
        Variable whose derivatives will be computed.
    wrt : str
        Variable with respect to which the derivatives will be computed.
    check_partials_data : dict of dicts of dicts
        First key:
            is the component name;
        Second key:
            is the (output, input) tuple of strings;
        Third key:
            is one of ['rel error', 'abs error', 'magnitude', 'J_fd', 'J_fwd', 'J_rev'];
        For 'rel error', 'abs error', 'magnitude' the value is: A tuple containing norms for
            forward - fd, adjoint - fd, forward - adjoint.
        For 'J_fd', 'J_fwd', 'J_rev' the value is: A numpy array representing the computed
            Jacobian for the three different methods of computation.
    title : str (Optional)
        Title for the plot
        If None, use the values of the arguments "of" and "wrt".
    jac_method : str (Optional)
        Method of computating Jacobian
        Is one of ['J_fwd', 'J_rev']. Optional, default is 'J_fwd'.
    tol : float (Optional)
        The tolerance, below which the two numbers are considered the same for
        plotting purposes.
    binary : bool (Optional)
        If true, the plot will only show the presence of a non-zero derivative, not the value.
        Otherwise, plot the value. Default is true.

    Returns
    -------
    matplotlib.figure.Figure
        The top level container for all the plot elements in the plot.
    array of matplotlib.axes.Axes objects
        An array of Axes objects, one for each of the three subplots created.

    Raises
    ------
    KeyError
        If one of the Jacobians is not available.
    """
    # Get the first item in the dict, which will be the model
    model_name = list(check_partials_data)[0]
    model_jacs = check_partials_data[model_name]
    key = (of, wrt)
    model_jac = model_jacs[key]

    # get finite difference arrays
    if 'J_fd' not in model_jac:
        msg = 'Jacobian "{}" not found.'
        raise KeyError(msg.format('J_fd'))
    fd_full = model_jac['J_fd']
    fd_full_flat = fd_full.flatten()  # for getting max min later
    if binary:
        fd_binary = fd_full.copy()
        fd_binary[np.nonzero(fd_binary)] = 1.0

    # get computed arrays
    if jac_method not in model_jac:
        msg = 'Jacobian "{}" not found.'
        raise KeyError(msg.format(jac_method))
    computed_full = model_jac[jac_method]
    computed_full_flat = computed_full.flatten()  # for getting max min later
    if binary:
        computed_binary = computed_full.copy()
        computed_binary[np.nonzero(computed_binary)] = 1.0

    # get plotting scales
    stacked = np.hstack((fd_full_flat, computed_full_flat))
    vmin = np.amin(stacked)
    vmax = np.amax(stacked)

    # basics of plot
    # Cannot do this sooner because of import and matplotlib
    #  backend issues when testing
    import matplotlib.pyplot as plt
    BINARY_CMP = plt.cm.gray
    NON_BINARY_CMP = plt.cm.viridis
    NON_BINARY_DIFF_CMP = plt.cm.RdBu
    fig, ax = plt.subplots(ncols=3, figsize=(12, 6))
    if title is None:
        title = str(key)
    plt.suptitle(title)

    # plot Jacobians
    if binary:
        ax[0].imshow(fd_binary.real, interpolation='none',
                     cmap=BINARY_CMP, aspect='auto')
        im_computed = ax[1].imshow(computed_binary.real, interpolation='none', cmap=BINARY_CMP,
                                   aspect='auto')
    else:
        ax[0].imshow(fd_full.real, interpolation='none', vmin=vmin, vmax=vmax, cmap=NON_BINARY_CMP,
                     aspect='auto')
        im_computed = ax[1].imshow(computed_full.real, interpolation='none', vmin=vmin, vmax=vmax,
                                   cmap=NON_BINARY_CMP, aspect='auto')
    ax[0].set_title('Approximated Jacobian')
    ax[1].set_title('User-Defined Jacobian')
    # Legend
    fig.colorbar(im_computed, orientation='horizontal',
                 ax=ax[0:2].ravel().tolist())

    # plot difference between computed and finite differenced
    diff = computed_full.real - fd_full.real
    diff_flat = diff.flatten()
    vmin = np.amin(diff_flat)
    vmax = np.amax(diff_flat)
    if vmax - vmin < tol:  # Do not want range to be too small
        vmin = -1 * tol
        vmax = tol
    color_limit = max(abs(vmin), abs(vmax))
    im_diff = ax[2].imshow(diff, interpolation='none', vmin=-color_limit, vmax=color_limit,
                           cmap=NON_BINARY_DIFF_CMP, aspect='auto')
    fig.colorbar(im_diff, orientation='horizontal', ax=ax[2], aspect=10)
    ax[2].set_title('Difference')

    plt.show()
    return fig, ax



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.units import convert_units


def units_setter(opt_meta, value):
    """
    Check and convert new units tuple into

    Parameters
    ----------
    opt_meta : dict
        Dictionary of entries for the option.
    value : any
        New value for the option.

    Returns
    -------
    any
        Post processed value to set into the option.
    """
    new_val, new_units = value
    old_val, units = opt_meta['val']

    converted_val = convert_units(new_val, new_units, units)
    return (converted_val, units)


class AviaryComp(om.ExplicitComponent):

    def setup(self):

        self.add_input('x', 3.0)
        self.add_output('y', 3.0)

    def initialize(self):
        self.options.declare('length', default=(12.0, 'inch'),
                             set_function=units_setter)

    def compute(self, inputs, outputs):
        length = self.options['length'][0]

        x = inputs['x']
        outputs['y'] = length * x


class Fakeviary(om.Group):

    def setup(self):
        self.add_subsystem('mass', AviaryComp())


class TestOptionsDictionaryUnits(unittest.TestCase):

    def test_simple(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('statics', Fakeviary())

        prob.model_options['*'] = {'length': (2.0, 'ft')}
        prob.setup()

        prob.run_model()

        y = prob.get_val('statics.mass.y')
        assert_near_equal(y, 72, 1e-6)


if __name__ == "__main__":
    unittest.main()



import sys
from functools import partial
import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal


class TestJaxUtils(unittest.TestCase):

    def test_jit_stub_no_jax(self):
        """Test the functionality of the jit stub."""
        from openmdao.utils.jax_utils import jit_stub as jit
        import numpy as np

        class TestClass(object):

            @partial(jit, static_argnums=(0,))
            def f(self, x):
                return np.sqrt(x)

        x = np.linspace(0, 10, 11)
        assert_near_equal(TestClass().f(x)**2, x)

    def test_jit_with_jax(self):
        """Make sure the jit stub test case also works with jax."""
        try:
            import jax.numpy as np
            from jax import jit
            import jaxlib
        except ImportError:
            self.skipTest('jax is not available but required for this test.')

        class TestClass(object):

            @partial(jit, static_argnums=(0,))
            def f(self, x):
                return np.sqrt(x)

        x = np.linspace(0, 10, 11)
        result = TestClass().f(x)
        assert_near_equal(result**2, x)
        self.assertIsInstance(result, jaxlib.xla_extension.ArrayImpl)

    def test_jax_component_option(self):
        """Test that the registration of jax-compatible components works."""
        import numpy as np

        if sys.version_info < (3, 9):
            self.skipTest('JaxExplicitComponent requires Python 3.9+')

        from openmdao.utils.jax_utils import jax
        if jax is None:
            self.skipTest('jax is not available.')

        class PowComp(om.JaxExplicitComponent):

            def initialize(self):
                self.options.declare('vec_size', types=(int,))
                self.options.declare('pow', types=(int,))

            def setup(self):
                n = self.options['vec_size']
                self.add_input('x', shape=(n,))
                self.add_output('f', shape=(n,))

                ar = np.arange(n, dtype=int)
                self.declare_partials(of='f', wrt='x', rows=ar, cols=ar)

            def get_self_statics(self):
                return (self.options['pow'],)

            def compute_primal(self, x):
                return x**self.options['pow']

        p = om.Problem()
        powcomp = p.model.add_subsystem('c', PowComp(derivs_method='jax', vec_size=11, pow=2))
        # deriv shape: [(11, 1, 11, 1)]
        p.setup(mode='fwd', force_alloc_complex=True)
        c_x = np.linspace(1, 10, 11)#.reshape(11, 1)
        p.set_val('c.x', c_x)
        p.run_model()
        assert_near_equal(p.get_val('c.f'), c_x ** 2)
        assert_check_partials(p.check_partials(method='cs', show_only_incorrect=False))

        p.set_val('c.x', c_x)
        powcomp.options['pow'] = 3  # change the option to verify that re-jit happens
        p.run_model()
        assert_near_equal(p.get_val('c.f'), c_x ** 3)
        assert_check_partials(p.check_partials(method='cs', show_only_incorrect=False))


if __name__ == '__main__':
    unittest.main()


"""
A module for OpenMDAO-specific warnings and associated functions.
"""

from functools import cmp_to_key
import inspect
import re
import sys
import io
import warnings

from openmdao import __version__ as om_version


__all__ = ['issue_warning', 'reset_warnings', 'reset_warning_registry', '_warn_simple_format',
           'OpenMDAOWarning', 'SetupWarning', 'DistributedComponentWarning', 'CaseRecorderWarning',
           'CacheWarning', 'PromotionWarning', 'UnusedOptionWarning', 'DerivativesWarning',
           'MPIWarning', 'UnitsWarning', 'SolverWarning', 'DriverWarning', 'OMDeprecationWarning',
           'OMInvalidCheckDerivativesOptionsWarning']


class OpenMDAOWarning(UserWarning):
    """
    Base class for all OpenMDAO warnings.
    """

    name = 'warn_openmdao'
    filter = 'always'


class SetupWarning(OpenMDAOWarning):
    """
    Warning class for warnings that occur during setup.
    """

    name = 'warn_setup'
    filter = 'always'


class PromotionWarning(SetupWarning):
    """
    Warning dealing with the promotion of an input or output.
    """

    name = 'warn_promotion'
    filter = 'always'


class UnitsWarning(SetupWarning):
    """
    Warning which is issued when unitless variable is connected to a variable with units.
    """

    name = 'warn_units'
    filter = 'always'


class DerivativesWarning(OpenMDAOWarning):
    """
    Warning issued when the approximated partials or coloring cannot be evaluated as expected.
    """

    name = 'warn_derivatives'
    filter = 'always'


class MPIWarning(SetupWarning):
    """
    Warning dealing with the availability of MPI.
    """

    name = 'warn_mpi'
    filter = 'always'


class DistributedComponentWarning(SetupWarning):
    """
    Warning specific to a distributed component.
    """

    name = 'warn_distributed_component'
    filter = 'always'


class SolverWarning(OpenMDAOWarning):
    """
    Warning base class for solver-related warnings.
    """

    name = 'warn_solver'
    filter = 'always'


class DriverWarning(OpenMDAOWarning):
    """
    Warning which is issued during the execution of a driver.
    """

    name = 'warn_driver'
    filter = 'always'


class UnusedOptionWarning(OpenMDAOWarning):
    """
    Warning dealing with an unnecessary option or argument being provided.
    """

    name = 'warn_unused_option'
    filter = 'always'


class CaseRecorderWarning(OpenMDAOWarning):
    """
    Warning pertaining to case recording and reading.
    """

    name = 'warn_case_recorder'
    filter = 'always'


class CacheWarning(OpenMDAOWarning):
    """
    Warning which is issued when the a cache is invalid and needs to be rebuilt.
    """

    name = 'warn_cache'
    filter = 'always'


class OMDeprecationWarning(OpenMDAOWarning):
    """
    An OpenMDAO-specific deprecation warning that is noisy by default, unlike the Python one.
    """

    name = 'warn_deprecation'
    filter = 'once'


class OMInvalidCheckDerivativesOptionsWarning(OpenMDAOWarning):
    """
    An OpenMDAO-specific warning that defaults to raising an exception.

    It tells the user that they are checking derivatives with the exact same method
    used to computer the derivatives, including all the settings.
    """

    name = 'warn_invalid_check_derivatives'
    filter = 'error'


_warnings = [_class for _, _class in
             inspect.getmembers(sys.modules[__name__], inspect.isclass)
             if issubclass(_class, Warning)]


# The order of warnings matters in the Python warnings module! See
#    https://docs.python.org/3/library/warnings.html#the-warnings-filter
# reset_warnings processes the warnings classes in the order returned by inspect.getmembers.
# The order is alphabetized. But this causes problems.
# To solve this problem, we sort the warnings classes so that parents come before their children
def _compare_class_hierarchy(class1, class2):
    if issubclass(class1, class2):
        return 1
    elif issubclass(class2, class1):
        return -1
    else:
        return 0


_warnings = sorted(_warnings, key=cmp_to_key(_compare_class_hierarchy))


def reset_warnings():
    """
    Apply the default warning filter actions for the OpenMDAO-specific warnings.

    This is necessary when testing the filters, because Python resets the default filters
    before running each test.
    """
    for w_class in _warnings:
        warnings.filterwarnings(w_class.filter, category=w_class)


def issue_warning(msg, prefix='', stacklevel=2, category=OpenMDAOWarning):
    """
    Display a warning with the desired stack level and optional prefix.

    Ex.

    >>> om.issue_warning('some warning message', prefix=self.pathname, category=om.SetupWarning)

    Parameters
    ----------
    msg : str
        The warning message.
    prefix : str
        An optional prefix to be prepended to the warning message (usually the system path).
    stacklevel : int
        Number of levels up the stack to identify as the warning location.
    category : class
        The class of warning to be issued.
    """
    old_format = warnings.formatwarning
    warnings.formatwarning = _warn_simple_format
    _msg = f'{prefix}: {msg}' if prefix else msg
    try:
        warnings.warn(_msg, category=category, stacklevel=stacklevel)
    finally:
        warnings.formatwarning = old_format


def _make_table(superclass=OpenMDAOWarning):
    """
    Generate a markdown table of the warning options.

    Returns
    -------
    str
        A string representation of a markdown table of the warning options.
    """
    s = io.StringIO()
    max_name_len = max([len(_class.__name__) for _class in _warnings])
    max_desc_len = max([len(' '.join(c.__doc__.split())) for c in _warnings])

    name_header = "Warning Class"
    desc_header = "Description"
    print(f'| {name_header:<{max_name_len}} | {desc_header:<{max_desc_len}} |', file=s)
    print(f'| {max_name_len * "-"} | {max_desc_len * "-"} |', file=s)

    for _class in _warnings:
        if isinstance(_class, superclass) or issubclass(_class, superclass):
            desc = ' '.join(_class.__doc__.split())
            print(f'| {_class.__name__:<{max_name_len}} | {desc:<{max_desc_len}} |', file=s)
    return s.getvalue()


def warn_deprecation(msg, expires=None):
    """
    Raise a warning and prints a deprecation message to stdout.

    Parameters
    ----------
    msg : str
        Message that will be printed to stdout.
    expires : str
        The version on OpenMDAO in which this deprecation expires, if any.
        If this deprecation is encountered and the OpenMDAO version is at least the given
        expires version, then an error is raised. This version should be in the form
        'major.minor.path', without any development/release suffix.

    Raises
    ------
    RuntimeError
        Raised if expires is provided and the current OpenMDAO version is greater than
        or equal to the expiration version.
    """
    # note, stack level 3 should take us back to original caller.
    if expires is not None:
        om_numeric = om_version.split('-')[0]
        om_version_tuple = tuple([int(s) for s in om_numeric.split('.')])
        expires_version_tuple = tuple([int(s) for s in expires.split('.')])
        if om_version_tuple >= expires_version_tuple:
            raise RuntimeError(f'Deprecation message expired in version {expires}')
    issue_warning(msg, stacklevel=3, category=OMDeprecationWarning)


def _warn_simple_format(message, category, filename, lineno, file=None, line=None):
    """
    Provide a warning format for OpenMDAO warnings.

    Parameters
    ----------
    message : str
        The warning message
    category : class
        The warning class being issued.
    filename : str
        The filename from which the warning is issued.
    lineno : int
        The line number from which the warning is issued.
    file : str
        Ignored in this implementation.
    line : str
        The line of code causing the warning (ignored in this implementation)

    Returns
    -------
    str
        A formatted warning message.

    """
    return f'{filename}:{lineno}: {category.__name__}:{message}\n'


class reset_warning_registry(object):
    """
    Context manager which archives & clears warning registry for duration of context.

    From https://bugs.python.org/file40031/reset_warning_registry.py
    (https://bugs.python.org/issue21724, https://github.com/python/cpython/issues/65923)

    Parameters
    ----------
    pattern : regex pattern
        Causes manager to only reset modules whose names match pattern. defaults to ``".*"``.

    Attributes
    ----------
    _pattern : regex pattern
        Causes manager to only reset modules whose names match this pattern. defaults to ``".*"``.
    """

    #: regexp for filtering which modules are reset
    _pattern = None

    #: dict mapping module name -> old registry contents
    _backup = None

    def __init__(self, pattern=None):
        """
        Initialize all attributes.
        """
        self._pattern = re.compile(pattern or ".*")

    def __enter__(self):
        """
        Enter the runtime context related to this object.

        Returns
        -------
        reset_warning_registry
            This context manager.

        """
        # archive and clear the __warningregistry__ key for all modules
        # that match the 'reset' pattern.
        pattern = self._pattern
        backup = self._backup = {}
        for name, mod in list(sys.modules.items()):
            if pattern.match(name):
                reg = getattr(mod, "__warningregistry__", None)
                if reg:
                    backup[name] = reg.copy()
                    reg.clear()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the runtime context related to this object.

        Parameters
        ----------
        exc_type : Exception class
            The type of the exception.
        exc_value : Exception instance
            The exception instance raised.
        traceback : regex pattern
            Traceback object.
        """
        # restore warning registry from backup
        modules = sys.modules
        backup = self._backup
        for name, content in backup.items():
            mod = modules.get(name)
            if mod is None:
                continue
            reg = getattr(mod, "__warningregistry__", None)
            if reg is None:
                setattr(mod, "__warningregistry__", content)
            else:
                reg.clear()
                reg.update(content)

        # clear all registry entries that we didn't archive
        pattern = self._pattern
        for name, mod in list(modules.items()):
            if pattern.match(name) and name not in backup:
                reg = getattr(mod, "__warningregistry__", None)
                if reg:
                    reg.clear()


# When we import OpenMDAO and load this module, set the default filters on these warnings.
reset_warnings()



import unittest

import numpy as np
from numpy.testing import assert_equal

from openmdao.utils.indexer import indexer, combine_ranges


class IndexerTestCase(unittest.TestCase):
    def test_int(self):
        ind = indexer[4]
        src = np.arange(10)

        assert_equal(ind(), 4)
        assert_equal(ind.as_array(), np.array([4]))
        assert_equal(ind.flat(), np.array([4]))

        assert_equal(src[ind()], 4)
        assert_equal(src[ind.as_array()], np.array([4]))
        assert_equal(src[ind.flat()], np.array([4]))

        ind.set_src_shape(src.shape)
        assert_equal(src[ind.shaped_array()], np.array([4]))
        assert_equal(ind.shaped_array(), np.array([4]))
        assert_equal(ind.indexed_src_shape, (1,))
        assert_equal(ind.min_src_dim, 1)

    def test_neg_int(self):
        ind = indexer[-4]
        src = np.arange(10)

        assert_equal(ind(), -4)
        assert_equal(ind.as_array(), np.array([-4]))
        assert_equal(ind.flat(), np.array([-4]))

        assert_equal(src[ind()], 6)
        assert_equal(src[ind.as_array()], np.array([6]))
        assert_equal(src[ind.flat()], np.array([6]))

        ind.set_src_shape(src.shape)
        assert_equal(ind.indexed_src_shape, (1,))
        assert_equal(ind.min_src_dim, 1)

        assert_equal(ind.shaped_array(), np.array([6]))
        assert_equal(src[ind.shaped_array()], np.array([6]))
        assert_equal(ind.shaped_instance()(), 6)

    def test_int_nonflat(self):
        ind = indexer[1]
        src = np.arange(9).reshape((3,3))

        assert_equal(ind(), 1)
        assert_equal(ind.as_array(), np.array([1]))
        assert_equal(ind.flat(), np.array([1]))

        assert_equal(src[ind()], np.array([3,4,5]))
        assert_equal(src[ind.as_array()], np.array([[3,4,5]]))
        assert_equal(src[ind.flat()], np.array([[3,4,5]]))

        ind.set_src_shape(src.shape)
        assert_equal(src[ind.shaped_array()], np.array([[3,4,5]]))
        assert_equal(ind.shaped_array(), np.array([1]))
        assert_equal(ind.indexed_src_shape, (3,))
        assert_equal(ind.min_src_dim, 1)

    def test_simple_noncontig_arr(self):
        ind = indexer[[5, 3, 7, 1]]
        src = np.arange(10)

        assert_equal(ind(), np.array([5,3,7,1]))
        assert_equal(ind.as_array(), np.array([5,3,7,1]))
        assert_equal(ind.flat(), np.array([5,3,7,1]))

        assert_equal(src[ind()], np.array([5,3,7,1]))
        assert_equal(src[ind.as_array()], np.array([5,3,7,1]))
        assert_equal(src[ind.flat()], np.array([5,3,7,1]))

        assert_equal(ind.min_src_dim, 1)

        ind.set_src_shape(src.shape)
        assert_equal(ind.indexed_src_shape, (4,))
        assert_equal(ind.shaped_instance()(), np.array([5,3,7,1]))
        assert_equal(ind.shaped_array(), np.array([5,3,7,1]))
        assert_equal(src[ind.shaped_array()], np.array([5,3,7,1]))

    def test_contiguous_arr(self):
        ind = indexer[[3, 4, 5]]
        src = np.arange(10)

        assert_equal(ind(), [3, 4, 5])
        assert_equal(ind.as_array(), np.array([3, 4, 5]))
        assert_equal(ind.flat(), [3, 4, 5])

        assert_equal(src[ind()], np.array([3, 4, 5]))
        assert_equal(src[ind.as_array()], np.array([3, 4, 5]))
        assert_equal(src[ind.flat()], np.array([3, 4, 5]))

        assert_equal(ind.min_src_dim, 1)

        ind.set_src_shape(src.shape)
        assert_equal(ind.indexed_src_shape, (3,))
        assert_equal(ind.shaped_instance()(), [3, 4, 5])
        assert_equal(src[ind.shaped_array()], np.array([3, 4, 5]))
        assert_equal(ind.shaped_array(), np.array([3, 4, 5]))

    def test_arr_to_slice_step2(self):
        ind = indexer[[1,3,5,7,9]]
        src = np.arange(10)

        assert_equal(ind(), [1,3,5,7,9])
        assert_equal(ind.flat(), [1,3,5,7,9])
        assert_equal(ind.as_array(), np.array([1,3,5,7,9]))

        assert_equal(src[ind()], np.array([1,3,5,7,9]))
        assert_equal(src[ind.as_array()], np.array([1,3,5,7,9]))
        assert_equal(src[ind.flat()], np.array([1,3,5,7,9]))

        ind.set_src_shape(src.shape)
        assert_equal(src[ind.shaped_array()], np.array([1,3,5,7,9]))
        assert_equal(ind.shaped_array(), np.array([1,3,5,7,9]))
        assert_equal(ind.indexed_src_shape, (5,))
        assert_equal(ind.min_src_dim, 1)

    def test_neg_arr(self):
        ind = indexer[[5, 3, 7, -1]]
        src = np.arange(10)

        assert_equal(ind(), np.array([5, 3, 7, -1]))
        assert_equal(ind.as_array(), np.array([5, 3, 7, -1]))
        assert_equal(ind.flat(), np.array([5, 3, 7, -1]))

        assert_equal(src[ind()], np.array([5,3,7,9]))
        assert_equal(src[ind.as_array()], np.array([5,3,7,9]))
        assert_equal(src[ind.flat()], np.array([5,3,7,9]))

        assert_equal(ind.indexed_src_shape, (4,))
        assert_equal(ind.min_src_dim, 1)

        try:
            ind.shaped_array()
        except Exception as err:
            self.assertEqual(str(err), "Can't get shaped array of [ 5  3  7 -1] because it has no source shape.")
        else:
            self.fail("Exception expected")

        ind.set_src_shape(src.shape)
        assert_equal(ind.shaped_array(), np.array([5,3,7,9]))
        assert_equal(src[ind.shaped_array()], np.array([5,3,7,9]))
        assert_equal(ind.shaped_instance()(), np.array([5,3,7,9]))

    def test_neg_arr2(self):
        ind = indexer[[-1, -3, -5]]
        src = np.arange(10)

        assert_equal(ind(), np.array([-1, -3, -5]))
        assert_equal(ind.as_array(), np.array([-1, -3, -5]))
        assert_equal(ind.flat(), np.array([-1, -3, -5]))

        assert_equal(src[ind()], np.array([9, 7, 5]))
        assert_equal(src[ind.as_array()], np.array([9, 7, 5]))
        assert_equal(src[ind.flat()], np.array([9, 7, 5]))

        assert_equal(ind.min_src_dim, 1)

        try:
            ind.shaped_array()
        except Exception as err:
            self.assertEqual(str(err), "Can't get shaped array of [-1 -3 -5] because it has no source shape.")
        else:
            self.fail("Exception expected")

        ind.set_src_shape(src.shape)
        assert_equal(ind.indexed_src_shape, (3,))
        assert_equal(ind.indexed_src_shape, (3,))
        assert_equal(ind.min_src_dim, 1)

        assert_equal(ind.shaped_instance()(), np.array([9, 7, 5]))
        assert_equal(ind.shaped_array(), np.array([9, 7, 5]))

    def test_full_slice(self):
        ind = indexer[:]
        src = np.arange(10)

        assert_equal(ind(), slice(None, None, 1))
        assert_equal(ind.flat(), slice(None, None, 1))

        assert_equal(src[ind()], np.arange(10, dtype=int))
        assert_equal(src[ind.flat()], np.arange(10, dtype=int))

        assert_equal(ind.min_src_dim, 1)

        with self.assertRaises(Exception) as cm:
            ind.indexed_src_shape
        self.assertEqual(cm.exception.args[0], "Can't get indexed_src_shape of slice(None, None, 1) because source shape is unknown.")
        with self.assertRaises(Exception) as cm:
            ind.as_array()
        self.assertEqual(cm.exception.args[0], "Can't get shaped array of slice(None, None, 1) because it has no source shape.")

        ind.set_src_shape(src.shape)

        assert_equal(ind.shaped_instance()(), slice(0, 10, 1))
        assert_equal(ind.as_array(), np.arange(10, dtype=int))
        assert_equal(ind.indexed_src_shape, (10,))
        assert_equal(ind.min_src_dim, 1)

    def test_neg_start_slice(self):
        ind = indexer[-3:-6:-1]
        src = np.arange(10)

        assert_equal(ind(), slice(-3, -6, -1))
        assert_equal(ind.flat(), slice(-3, -6, -1))

        assert_equal(src[ind()], np.array([7, 6, 5]))
        assert_equal(src[ind.flat()], np.array([7, 6, 5]))

        assert_equal(ind.min_src_dim, 1)

        ind.set_src_shape(src.shape)

        assert_equal(ind.shaped_instance()(), slice(7, 4, -1))
        assert_equal(ind.as_array(), np.array([7, 6, 5]))
        assert_equal(ind.indexed_src_shape, (3,))
        assert_equal(ind.min_src_dim, 1)

    def test_none_start_slice(self):
        ind = indexer[:5]
        src = np.arange(10)

        assert_equal(ind(), slice(None, 5, 1))
        assert_equal(ind.flat(), slice(None, 5, 1))

        assert_equal(src[ind()], np.array([0, 1, 2, 3, 4]))
        assert_equal(src[ind.flat()], np.array([0, 1, 2, 3, 4]))

        assert_equal(ind.min_src_dim, 1)

        ind.set_src_shape(src.shape)
        assert_equal(ind.as_array(), np.array([0, 1, 2, 3, 4]))
        assert_equal(src[ind.as_array()], np.array([0, 1, 2, 3, 4]))

        assert_equal(ind.shaped_instance()(), slice(None, 5, 1))
        assert_equal(ind.indexed_src_shape, (5,))
        assert_equal(ind.min_src_dim, 1)

    def test_none_stop_slice(self):
        ind = indexer[3:]
        src = np.arange(10)

        assert_equal(ind(), slice(3, None, 1))
        assert_equal(ind.flat(), slice(3, None, 1))

        assert_equal(src[ind()], np.array([3, 4, 5, 6, 7, 8, 9]))
        assert_equal(src[ind.flat()], np.array([3, 4, 5, 6, 7, 8, 9]))

        assert_equal(ind.min_src_dim, 1)

        with self.assertRaises(Exception) as cm:
            ind.as_array()
        self.assertEqual(cm.exception.args[0], "Can't get shaped array of slice(3, None, 1) because it has no source shape.")

        ind.set_src_shape(src.shape)

        assert_equal(ind.shaped_instance()(), slice(3, 10, 1))
        assert_equal(ind.as_array(), np.array([3, 4, 5, 6, 7, 8, 9]))
        assert_equal(ind.indexed_src_shape, (7,))
        assert_equal(ind.min_src_dim, 1)

    def test_slice_neg_step(self):
        ind = indexer[6:2:-1]
        src = np.arange(10)

        assert_equal(ind(), slice(6, 2, -1))
        assert_equal(ind.flat(), slice(6, 2, -1))

        assert_equal(src[ind()], np.array([6, 5, 4, 3]))
        assert_equal(src[ind.flat()], np.array([6, 5, 4, 3]))

        assert_equal(ind.min_src_dim, 1)

        ind.set_src_shape(src.shape)

        assert_equal(ind.as_array(), np.array([6, 5, 4, 3]))
        assert_equal(src[ind.as_array()], np.array([6, 5, 4, 3]))
        assert_equal(src[ind.shaped_array()], np.array([6, 5, 4, 3]))
        assert_equal(ind.shaped_instance()(), slice(6, 2, -1))
        assert_equal(ind.indexed_src_shape, (4,))
        assert_equal(ind.min_src_dim, 1)

    def test_rev_slice(self):
        ind = indexer[::-1]
        src = np.arange(10)

        assert_equal(ind(), slice(None, None, -1))
        assert_equal(ind.flat(), slice(None, None, -1))

        assert_equal(src[ind()], np.array([9, 8, 7, 6, 5, 4, 3, 2, 1, 0]))
        assert_equal(src[ind.flat()], np.array([9, 8, 7, 6, 5, 4, 3, 2, 1, 0]))

        assert_equal(ind.min_src_dim, 1)

        ind.set_src_shape(src.shape)

        assert_equal(ind.as_array(), np.array([9, 8, 7, 6, 5, 4, 3, 2, 1, 0]))
        assert_equal(src[ind.as_array()], np.array([9, 8, 7, 6, 5, 4, 3, 2, 1, 0]))
        assert_equal(src[ind.shaped_array()], np.array([9, 8, 7, 6, 5, 4, 3, 2, 1, 0]))
        assert_equal(ind.shaped_instance()(), slice(None, None, -1))
        assert_equal(ind.indexed_src_shape, (10,))
        assert_equal(ind.min_src_dim, 1)

    def test_rev_slice2D(self):
        ind = indexer[::-1]
        src = np.arange(4).reshape((2,2))

        assert_equal(ind(), slice(None, None, -1))
        assert_equal(ind.flat(), slice(None, None, -1))

        expected = src[::-1]

        assert_equal(src[ind()], expected)
        assert_equal(src[ind.flat()], expected)

        assert_equal(ind.min_src_dim, 1)

        ind.set_src_shape(src.shape)

        assert_equal(ind.as_array(), expected.flatten())
        assert_equal(src[ind()], expected)
        assert_equal(ind.shaped_instance()(), slice(None, None, -1))
        assert_equal(ind.indexed_src_shape, (2,2))
        assert_equal(ind.min_src_dim, 1)

class IndexerMultiDimTestCase(unittest.TestCase):
    def test_multi_slice(self):
        ind = indexer[:,:,:]
        src = np.arange(27).reshape((3,3,3))

        assert_equal(ind(), (slice(None, None, 1), slice(None, None, 1), slice(None, None, 1)))
        assert_equal(src[ind()], src)

        with self.assertRaises(Exception) as cm:
            ind.flat()
        self.assertEqual(cm.exception.args[0], "Can't get shaped array of (slice(None, None, None), slice(None, None, None), slice(None, None, None)) because it has no source shape.")

        with self.assertRaises(Exception) as cm:
            ind.indexed_src_shape
        self.assertEqual(cm.exception.args[0], "Can't get indexed_src_shape of (slice(None, None, None), slice(None, None, None), slice(None, None, None)) because source shape is unknown.")
        with self.assertRaises(Exception) as cm:
            ind.as_array()
        self.assertEqual(cm.exception.args[0], "Can't determine extent of array because source shape is not known.")

        ind.set_src_shape(src.shape)

        assert_equal(src.flat[ind.flat()], np.arange(27, dtype=int))

        assert_equal(ind.shaped_instance()(), (slice(0, 3, 1), slice(0, 3, 1), slice(0, 3, 1)))
        assert_equal(ind.as_array(flat=False), src)
        assert_equal(ind.as_array(), src.flat[:])
        assert_equal(ind.indexed_src_shape, (3,3,3))
        assert_equal(ind.min_src_dim, 3)

    def test_flat_slice_into_nd_source(self):
        ind = indexer[1:]
        src = np.arange(27).reshape((3,3,3))
        ind.set_src_shape(src.shape)

        assert_equal(ind(), slice(1, None, 1))

        expected = np.stack([np.arange(9, 18).reshape((3, 3)), np.arange(18, 27).reshape((3, 3))])

        assert_equal(expected, src[1:])
        assert_equal(src[ind()], expected)

        shaped_ind = ind.shaped_instance()

        assert_equal(shaped_ind.as_array(flat=True), expected.ravel())

        expected_nonflat = (np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2]),
                            np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 0, 0, 0, 1, 1, 1, 2, 2, 2]),
                            np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2]))

        assert_equal(shaped_ind.as_array(flat=False), expected_nonflat)

    def test_slice_neg(self):
        ind = indexer[:-1,:,:2]
        src = np.arange(27).reshape((3,3,3))

        assert_equal(ind(), (slice(None, -1, 1), slice(None, None, 1), slice(None, 2, 1)))
        assert_equal(src[ind()], src[:-1,:,:2])

        with self.assertRaises(Exception) as cm:
            ind.indexed_src_shape
        self.assertEqual(cm.exception.args[0], "Can't get indexed_src_shape of (slice(None, -1, None), slice(None, None, None), slice(None, 2, None)) because source shape is unknown.")
        with self.assertRaises(Exception) as cm:
            ind.as_array()
        self.assertEqual(cm.exception.args[0], "Can't determine extent of array because source shape is not known.")

        ind.set_src_shape(src.shape)

        assert_equal(ind.shaped_instance()(), (slice(0, 2, 1), slice(0, 3, 1), slice(None, 2, 1)))
        assert_equal(ind.as_array(), np.arange(27, dtype=np.int32).reshape((3,3,3))[:-1,:,:2].ravel())
        assert_equal(ind.as_array(flat=False), np.arange(27, dtype=np.int32).reshape((3,3,3))[:-1,:,:2])
        assert_equal(ind.indexed_src_shape, (2,3,2))
        assert_equal(ind.min_src_dim, 3)

    def test_mult_arr(self):
        src = np.arange(27).reshape((3,3,3))
        ind = indexer[[0,2], :, [1,2]]

        assert_equal(ind(), ([0,2], slice(None, None, 1), [1,2]))
        with self.assertRaises(Exception) as cm:
            ind.indexed_src_shape
        self.assertEqual(cm.exception.args[0], "Can't get indexed_src_shape of ([0, 2], slice(None, None, None), [1, 2]) because source shape is unknown.")
        with self.assertRaises(Exception) as cm:
            ind.as_array()
        self.assertEqual(cm.exception.args[0], "Can't determine extent of array because source shape is not known.")

        ind.set_src_shape(src.shape)

        assert_equal(ind.shaped_instance()(), ([0, 2], slice(0, 3, 1), [1, 2]))
        assert_equal(ind.as_array(), np.arange(27, dtype=np.int32).reshape((3,3,3))[[0,2], :, [1,2]].ravel())
        assert_equal(ind.as_array(flat=False), np.arange(27, dtype=np.int32).reshape((3,3,3))[[0,2], :, [1,2]])
        assert_equal(ind.indexed_src_shape, (2,3))
        assert_equal(ind.min_src_dim, 3)

    def test_mult_arr_shaped(self):
        src = np.arange(9).reshape((3,3))
        ind = indexer[[[0,0],[2,2]], [[0,2], [0, 2]]]  # shoud get a 2x2 array with corner values of the 3x3

        assert_equal(ind(), ([[0,0],[2,2]], [[0,2], [0, 2]]))
        with self.assertRaises(Exception) as cm:
            ind.indexed_src_shape
        self.assertEqual(cm.exception.args[0], "Can't get indexed_src_shape of ([[0, 0], [2, 2]], [[0, 2], [0, 2]]) because source shape is unknown.")
        with self.assertRaises(Exception) as cm:
            ind.as_array()
        self.assertEqual(cm.exception.args[0], "Can't determine extent of array because source shape is not known.")

        ind.set_src_shape(src.shape)

        assert_equal(ind.shaped_instance()(), ([[0,0],[2,2]], [[0,2], [0, 2]]))
        assert_equal(ind.as_array(), np.arange(9, dtype=np.int32).reshape((3,3))[[[0,0],[2,2]], [[0,2], [0, 2]]].ravel())
        assert_equal(ind.as_array(flat=False), np.arange(9, dtype=np.int32).reshape((3,3))[[[0,0],[2,2]], [[0,2], [0, 2]]])
        assert_equal(ind.indexed_src_shape, (2,2))
        assert_equal(ind.min_src_dim, 2)

    def test_flat_shaped_src_inds(self):
        src = np.arange(24).reshape((8,3))  # 2D source
        ind = indexer([1,3,5,4,22, -4, 11, 3], flat_src=True)

        assert_equal(ind(), [1,3,5,4,22, -4, 11, 3])

        ind.set_src_shape(src.shape)

        assert_equal(ind.shaped_array(), np.array([1,3,5,4,22, 20, 11, 3]))
        assert_equal(ind.as_array(), np.array([1,3,5,4,22, -4, 11, 3]))
        assert_equal(ind.as_array(flat=False), np.array([1,3,5,4,22, -4, 11, 3]))
        assert_equal(ind.indexed_src_shape, (8,))
        assert_equal(ind.min_src_dim, 1)


class TestCombineRanges(unittest.TestCase):

    def test_empty(self):
        ranges = []
        result = combine_ranges(ranges)
        self.assertEqual(result, [])

    def test_single_range(self):
        ranges = [(1, 5)]
        result = combine_ranges(ranges)
        self.assertEqual(result, [(1, 5)])

    def test_contig_ranges(self):
        ranges = [(1, 5), (5, 10), (10, 15)]
        result = combine_ranges(ranges)
        self.assertEqual(result, [(1, 15)])

    def test_non_overlapping_ranges(self):
        ranges = [(1, 5), (6, 10), (11, 15)]
        result = combine_ranges(ranges)
        self.assertEqual(result, [(1, 5), (6, 10), (11, 15)])

    def test_mixed_ranges(self):
        ranges = [(1, 5), (5, 10), (11, 15)]
        result = combine_ranges(ranges)
        self.assertEqual(result, [(1, 10), (11, 15)])


if __name__ == '__main__':
    unittest.main()



"""Test the Broyden nonlinear solver. """

import unittest

import numpy as np

import openmdao.api as om

from openmdao.test_suite.components.double_sellar import DoubleSellar
from openmdao.test_suite.components.implicit_newton_linesearch import ImplCompTwoStates
from openmdao.test_suite.components.sellar import SellarStateConnection, SellarDerivatives, \
     SellarDis1withDerivatives, SellarDis2withDerivatives
from openmdao.test_suite.scripts.circuit_analysis import Circuit
from openmdao.utils.assert_utils import assert_near_equal, assert_warning

try:
    from openmdao.vectors.petsc_vector import PETScVector
except ImportError:
    PETScVector = None


class VectorEquation(om.ImplicitComponent):
    """Equation with 5 states in a single vector. Should converge to x=[0,0,0,0,0]"""

    def setup(self):
        self.add_input('c', 0.01)
        self.add_output('x', np.ones((5, )))

    def apply_nonlinear(self, inputs, outputs, residuals):
        c = inputs['c']
        x = outputs['x']

        d = np.array([3, 2, 1.5, 1, 0.5])

        residuals['x'] = -d*x - c*x**3


class MixedEquation(om.ImplicitComponent):
    """Equation with 5 states split between 3 vars. Should converge to x=[0,0,0,0,0]"""

    def setup(self):
        self.add_input('c', 0.01)
        self.add_output('x12', np.ones((2, )))
        self.add_output('x3', 1.0)
        self.add_output('x45', np.ones((2, )))

        self.declare_partials(of=['x12', 'x3', 'x45'], wrt='c')
        self.declare_partials(of='x12', wrt='x12', rows=np.arange(2), cols=np.arange(2),
                              val=-np.array([3.0, 2]))
        self.declare_partials(of='x3', wrt='x3', rows=np.arange(1), cols=np.arange(1),
                              val=-np.array([1.5]))
        self.declare_partials(of='x45', wrt='x45', rows=np.arange(2), cols=np.arange(2),
                              val=-np.array([1, 0.5]))

    def apply_nonlinear(self, inputs, outputs, residuals):
        c = inputs['c']
        x = np.empty((5, ))
        x[:2] = outputs['x12']
        x[2] = outputs['x3'].item()
        x[3:] = outputs['x45']

        d = np.array([3, 2, 1.5, 1, 0.5])

        res = -d*x - c*x**3
        residuals['x12'] = res[:2]
        residuals['x3'] = res[2]
        residuals['x45'] = res[3:]

    def linearize(self, inputs, outputs, jacobian):
        x12 = outputs['x12']
        x3 = outputs['x3']
        x45 = outputs['x45']

        jacobian['x12', 'c'] = -3.0 * x12**2
        jacobian['x3', 'c'] = -3.0 * x3**2
        jacobian['x45', 'c'] = -3.0 * x45**2


class SpedicatoHuang(om.ImplicitComponent):

    cite = """
           @article{spedicato_hwang,
           author = {E. Spedicato, Z. Huang},
           title = {Numerical experience with newton-like methods for nonlinear algebraic systems},
           journal = {Computing},
           voluem = {86},
           year = {1997},
           }
           """

    def setup(self):

        self.n = 3

        self.add_input('x', np.array([0, 20]))
        self.add_output('y', 10.0*np.ones((self.n, )))

        self.declare_partials(of='y', wrt=['x', 'y'])

    def apply_nonlinear(self, inputs, outputs, residuals):
        x = inputs['x']
        y = outputs['y']
        n = self.n

        residuals['y'][0] = y[0] + y[1] + x[0] + .25*(y[1] - x[0])**2
        residuals['y'][n-1] = y[n-1] + x[1] + y[n-2] + .25*(x[1] - y[n-2])**2
        for j in np.arange(1, n-1):
            residuals['y'][j] = y[j] + y[j+1] + y[j-1] + .25*(y[j+1] - y[j-1])**2

    def linearize(self, inputs, outputs, jacobian):
        x = inputs['x']
        y = outputs['y']
        n = self.n

        jacobian['y', 'x'][0, 0] = 1.0 - .5*(y[1] - x[0])
        jacobian['y', 'y'][0, 0] = 1.0
        jacobian['y', 'y'][0, 1] = 1.0 + .5*(y[1] - x[0])

        jacobian['y', 'x'][n-1, 1] = 1.0 + .5*(x[1] - y[n-2])
        jacobian['y', 'y'][n-1, n-1] = 1.0
        jacobian['y', 'y'][n-1, n-2] = 1.0 - .5*(x[1] - y[n-2])

        for j in np.arange(1, n-1):
            jacobian['y', 'y'][j, j-1] = 1.0 - .5*(y[j+1] - y[j-1])
            jacobian['y', 'y'][j, j] = 1.0
            jacobian['y', 'y'][j, j+1] = 1.0 + .5*(y[j+1] - y[j-1])


class TestBryoden(unittest.TestCase):

    def test_reraise_error(self):

        prob = om.Problem(model=DoubleSellar())
        model = prob.model

        g1 = model.g1
        g1.nonlinear_solver = om.BroydenSolver()
        g1.nonlinear_solver.options['maxiter'] = 1
        g1.nonlinear_solver.options['err_on_non_converge'] = True
        g1.linear_solver = om.DirectSolver(assemble_jac=True)

        g2 = model.g2
        g2.nonlinear_solver = om.BroydenSolver()
        g2.nonlinear_solver.options['maxiter'] = 1
        g2.nonlinear_solver.options['err_on_non_converge'] = True
        g2.linear_solver = om.DirectSolver(assemble_jac=True)

        model.nonlinear_solver = om.BroydenSolver()
        model.linear_solver = om.DirectSolver(assemble_jac=True)
        model.nonlinear_solver.options['err_on_non_converge'] = True
        model.nonlinear_solver.options['reraise_child_analysiserror'] = True

        prob.setup()

        with self.assertRaises(om.AnalysisError) as context:
            prob.run_model()

        msg = "Solver 'NL: BROYDEN' on system 'g1' failed to converge in 1 iterations."
        self.assertEqual(str(context.exception), msg)

    def test_error_badname(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())

        prob.setup()

        model.nonlinear_solver.options['state_vars'] = ['junk']

        with self.assertRaises(ValueError) as context:
            prob.run_model()

        msg = "BroydenSolver in <model> <class SellarStateConnection>: The following variable names were not found: junk"
        self.assertEqual(str(context.exception), msg)

    def test_error_need_direct_solver(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                           linear_solver=om.LinearRunOnce())

        prob.setup()

        with self.assertRaises(ValueError) as context:
            prob.run_model()

        msg = "BroydenSolver in <model> <class SellarStateConnection>: Linear solver must be DirectSolver when solving the full model."
        self.assertEqual(str(context.exception), msg)

    def test_simple_sellar(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())

        prob.setup()

        model.nonlinear_solver.options['state_vars'] = ['state_eq.y2_command']
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.run_model()

        assert_near_equal(prob['y1'], 25.58830273, .00001)
        assert_near_equal(prob['state_eq.y2_command'], 12.05848819, .00001)

    def test_simple_sellar_cycle(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarDerivatives(nonlinear_solver=om.BroydenSolver(),
                                               linear_solver=om.LinearRunOnce())

        prob.setup()

        model.nonlinear_solver.options['state_vars'] = ['y1']
        model.nonlinear_solver.options['compute_jacobian'] = True

        prob.set_solver_print(level=2)

        prob.run_model()

        assert_near_equal(prob['y1'], 25.58830273, .00001)
        assert_near_equal(prob['y2'], 12.05848819, .00001)

    def test_sellar_state_connection_fd_system(self):
        # Sellar model closes loop with state connection instead of a cycle.
        # This test is just fd.
        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())
        prob.model.approx_totals(method='fd')

        prob.setup()

        model.nonlinear_solver.options['state_vars'] = ['state_eq.y2_command']
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.run_model()

        assert_near_equal(prob.get_val('y1'), 25.58830273, .00001)
        assert_near_equal(prob.get_val('state_eq.y2_command'), 12.05848819, .00001)

        # Make sure we aren't iterating like crazy
        self.assertLess(prob.model.nonlinear_solver._iter_count, 6)

    def test_vector(self):
        # Testing Broyden on a 5 state single vector case.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('c', 0.01))
        model.add_subsystem('vec', VectorEquation())

        model.connect('p1.c', 'vec.c')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['vec.x']
        model.nonlinear_solver.options['maxiter'] = 15
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.setup()

        prob.run_model()

        assert_near_equal(prob['vec.x'], np.zeros((5, )), 1e-6)

    def test_mixed(self):
        # Testing Broyden on a 5 state case split among 3 vars.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('c', 0.01))
        model.add_subsystem('mixed', MixedEquation())

        model.connect('p1.c', 'mixed.c')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['mixed.x12', 'mixed.x3', 'mixed.x45']
        model.nonlinear_solver.options['maxiter'] = 15
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.setup()

        prob.run_model()

        assert_near_equal(prob['mixed.x12'], np.zeros((2, )), 1e-6)
        assert_near_equal(prob['mixed.x3'], 0.0, 1e-6)
        assert_near_equal(prob['mixed.x45'], np.zeros((2, )), 1e-6)

    def test_missing_state_warning(self):
        # Testing Broyden on a 5 state case split among 3 vars.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('c', 0.01))
        model.add_subsystem('mixed', MixedEquation())

        model.connect('p1.c', 'mixed.c')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['mixed.x12']
        model.nonlinear_solver.options['maxiter'] = 15
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.setup()

        msg = "The following states are not covered by a solver, and may have been " \
              "omitted from the BroydenSolver 'state_vars': mixed.x3, mixed.x45"

        with assert_warning(UserWarning, msg):
            prob.run_model()

        # Try again with promoted names.
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('c', 0.01))
        model.add_subsystem('mixed', MixedEquation(), promotes=['*'])

        model.connect('p1.c', 'c')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['x12']
        model.nonlinear_solver.options['maxiter'] = 15
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.setup()

        msg = "The following states are not covered by a solver, and may have been " \
              "omitted from the BroydenSolver 'state_vars': x3, x45"

        with assert_warning(UserWarning, msg):
            prob.run_model()

    def test_mixed_promoted_vars(self):
        # Testing Broyden on a 5 state case split among 3 vars.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('c', 0.01))
        model.add_subsystem('mixed', MixedEquation(), promotes_outputs=['x12', 'x3', 'x45'])

        model.connect('p1.c', 'mixed.c')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['x12', 'x3', 'x45']
        model.nonlinear_solver.options['maxiter'] = 15
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.setup()

        prob.run_model()

        assert_near_equal(prob['x12'], np.zeros((2, )), 1e-6)
        assert_near_equal(prob['x3'], 0.0, 1e-6)
        assert_near_equal(prob['x45'], np.zeros((2, )), 1e-6)

    def test_mixed_jacobian(self):
        # Testing Broyden on a 5 state case split among 3 vars.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('c', 0.01))
        model.add_subsystem('mixed', MixedEquation())

        model.connect('p1.c', 'mixed.c')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['mixed.x12', 'mixed.x3', 'mixed.x45']
        model.nonlinear_solver.options['maxiter'] = 15
        model.nonlinear_solver.linear_solver = om.DirectSolver()

        prob.setup()

        prob.run_model()

        assert_near_equal(prob['mixed.x12'], np.zeros((2, )), 1e-6)
        assert_near_equal(prob['mixed.x3'], 0.0, 1e-6)
        assert_near_equal(prob['mixed.x45'], np.zeros((2, )), 1e-6)

        # Normally takes about 13 iters, but takes around 4 if you calculate an initial
        # Jacobian.
        self.assertTrue(model.nonlinear_solver._iter_count < 6)

    def test_simple_sellar_jacobian(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())

        prob.setup()

        model.nonlinear_solver.options['state_vars'] = ['state_eq.y2_command']
        model.nonlinear_solver.linear_solver = om.DirectSolver(assemble_jac=False)

        prob.run_model()

        assert_near_equal(prob['y1'], 25.58830273, .00001)
        assert_near_equal(prob['state_eq.y2_command'], 12.05848819, .00001)

        # Normally takes about 4 iters, but takes around 3 if you calculate an initial
        # Jacobian.
        self.assertTrue(model.nonlinear_solver._iter_count < 4)

    def test_simple_sellar_jacobian_assembled(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())

        prob.setup()

        model.nonlinear_solver.linear_solver = om.DirectSolver(assemble_jac=True)

        prob.run_model()

        assert_near_equal(prob['y1'], 25.58830273, .00001)
        assert_near_equal(prob['state_eq.y2_command'], 12.05848819, .00001)

        # Normally takes about 4 iters, but takes around 3 if you calculate an initial
        # Jacobian.
        self.assertTrue(model.nonlinear_solver._iter_count < 4)

    def test_simple_sellar_jacobian_assembled_dense(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())

        prob.setup()

        model.options['assembled_jac_type'] = 'dense'
        model.nonlinear_solver.linear_solver = om.DirectSolver(assemble_jac=True)

        prob.run_model()

        assert_near_equal(prob['y1'], 25.58830273, .00001)
        assert_near_equal(prob['state_eq.y2_command'], 12.05848819, .00001)

        # Normally takes about 4 iters, but takes around 3 if you calculate an initial
        # Jacobian.
        self.assertTrue(model.nonlinear_solver._iter_count < 4)

    def test_simple_sellar_full(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())

        prob.setup()

        model.nonlinear_solver.linear_solver = om.DirectSolver()
        model.nonlinear_solver.options['compute_jacobian'] = False

        prob.run_model()

        assert_near_equal(prob['y1'], 25.58830273, .00001)
        assert_near_equal(prob['state_eq.y2_command'], 12.05848819, .00001)

        # Normally takes about 5 iters, but takes around 4 if you calculate an initial
        # Jacobian.
        self.assertTrue(model.nonlinear_solver._iter_count < 6)

    def test_simple_sellar_full_jacobian(self):
        # Test top level Sellar (i.e., not grouped).

        prob = om.Problem()
        model = prob.model = SellarStateConnection(nonlinear_solver=om.BroydenSolver(),
                                                   linear_solver=om.LinearRunOnce())

        prob.setup()

        model.nonlinear_solver.linear_solver = om.DirectSolver()

        prob.run_model()

        assert_near_equal(prob['y1'], 25.58830273, .00001)
        assert_near_equal(prob['state_eq.y2_command'], 12.05848819, .00001)

        # Normally takes about 5 iters, but takes around 4 if you calculate an initial
        # Jacobian.
        self.assertTrue(model.nonlinear_solver._iter_count < 5)

    def test_jacobian_update_converge_limit(self):
        # This model needs jacobian updates to converge.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', np.array([0, 20.0])))
        model.add_subsystem('comp', SpedicatoHuang())

        model.connect('p1.x', 'comp.x')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['comp.y']
        model.nonlinear_solver.options['maxiter'] = 20
        model.nonlinear_solver.options['max_converge_failures'] = 1
        model.nonlinear_solver.options['diverge_limit'] = np.inf
        model.nonlinear_solver.linear_solver = om.DirectSolver()

        prob.setup()

        prob.set_solver_print(level=2)
        prob.run_model()

        assert_near_equal(prob['comp.y'], np.array([-36.26230985,  10.20857237, -54.17658612]), 1e-6)

    def test_jacobian_update_diverge_limit(self):
        # This model needs jacobian updates to converge.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', np.array([0, 20.0])))
        model.add_subsystem('comp', SpedicatoHuang())

        model.connect('p1.x', 'comp.x')

        model.nonlinear_solver = om.BroydenSolver()
        model.nonlinear_solver.options['state_vars'] = ['comp.y']
        model.nonlinear_solver.options['maxiter'] = 20
        model.nonlinear_solver.options['diverge_limit'] = 0.5
        model.nonlinear_solver.linear_solver = om.DirectSolver()

        prob.setup()

        prob.set_solver_print(level=2)
        prob.run_model()

        assert_near_equal(prob['comp.y'], np.array([-36.26230985,  10.20857237, -54.17658612]), 1e-6)

    def test_backtracking(self):
        top = om.Problem()
        top.model.add_subsystem('px', om.IndepVarComp('x', 1.0))
        top.model.add_subsystem('comp', ImplCompTwoStates())
        top.model.connect('px.x', 'comp.x')

        top.model.nonlinear_solver = om.BroydenSolver()
        top.model.nonlinear_solver.options['maxiter'] = 25
        top.model.nonlinear_solver.options['diverge_limit'] = 0.5
        top.model.nonlinear_solver.options['state_vars'] = ['comp.y', 'comp.z']

        top.model.linear_solver = om.DirectSolver()

        top.setup()

        top.set_solver_print(level=2)
        # Test lower bound: should go to the lower bound and stall
        top['px.x'] = 2.0
        top['comp.y'] = 0.0
        top['comp.z'] = 1.6
        top.run_model()
        assert_near_equal(top['comp.z'], 1.5, 1e-8)

        # Test upper bound: should go to the upper bound and stall
        top['px.x'] = 0.5
        top['comp.y'] = 0.0
        top['comp.z'] = 2.4
        top.run_model()
        assert_near_equal(top['comp.z'], 2.5, 1e-8)

    def test_cs_around_broyden(self):
        # Basic sellar test.

        prob = om.Problem()
        model = prob.model
        sub = model.add_subsystem('sub', om.Group(), promotes=['*'])

        model.add_subsystem('px', om.IndepVarComp('x', 1.0), promotes=['x'])
        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])), promotes=['z'])

        sub.add_subsystem('d1', SellarDis1withDerivatives(), promotes=['x', 'z', 'y1', 'y2'])
        sub.add_subsystem('d2', SellarDis2withDerivatives(), promotes=['z', 'y1', 'y2'])

        model.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                z=np.array([0.0, 0.0]), x=0.0),
                            promotes=['obj', 'x', 'z', 'y1', 'y2'])

        model.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'),
                            promotes=['con1', 'y1'])
        model.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'),
                            promotes=['con2', 'y2'])

        sub.nonlinear_solver = om.BroydenSolver()
        sub.linear_solver = om.DirectSolver()
        model.linear_solver = om.DirectSolver()

        prob.model.add_design_var('x', lower=-100, upper=100)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.model.add_objective('obj')
        prob.model.add_constraint('con1', upper=0.0)
        prob.model.add_constraint('con2', upper=0.0)

        prob.setup(check=False, force_alloc_complex=True)
        prob.set_solver_print(level=2)

        prob.run_model()

        totals = prob.check_totals(method='cs', out_stream=None)

        for key, val in totals.items():
            assert_near_equal(val['rel error'][0], 0.0, 1e-6)

    def test_cs_around_broyden_compute_jac(self):
        # Basic sellar test.

        prob = om.Problem()
        model = prob.model
        sub = model.add_subsystem('sub', om.Group(), promotes=['*'])

        model.add_subsystem('px', om.IndepVarComp('x', 1.0), promotes=['x'])
        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])), promotes=['z'])

        sub.add_subsystem('d1', SellarDis1withDerivatives(), promotes=['x', 'z', 'y1', 'y2'])
        sub.add_subsystem('d2', SellarDis2withDerivatives(), promotes=['z', 'y1', 'y2'])

        model.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                z=np.array([0.0, 0.0]), x=0.0),
                            promotes=['obj', 'x', 'z', 'y1', 'y2'])

        model.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'), promotes=['con1', 'y1'])
        model.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'), promotes=['con2', 'y2'])

        sub.nonlinear_solver = om.BroydenSolver()
        sub.linear_solver = om.DirectSolver(assemble_jac=False)
        model.linear_solver = om.DirectSolver(assemble_jac=False)

        prob.model.add_design_var('x', lower=-100, upper=100)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.model.add_objective('obj')
        prob.model.add_constraint('con1', upper=0.0)
        prob.model.add_constraint('con2', upper=0.0)

        prob.setup(check=False, force_alloc_complex=True)
        prob.set_solver_print(level=0)

        prob.run_model()

        sub.nonlinear_solver.options['compute_jacobian'] = True

        totals = prob.check_totals(method='cs', out_stream=None)

        for key, val in totals.items():
            assert_near_equal(val['rel error'][0], 0.0, 1e-6)

    def test_cs_around_broyden_compute_jac_dense(self):
        # Basic sellar test.

        prob = om.Problem()
        model = prob.model
        sub = model.add_subsystem('sub', om.Group(), promotes=['*'])

        model.add_subsystem('px', om.IndepVarComp('x', 1.0), promotes=['x'])
        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])), promotes=['z'])

        sub.add_subsystem('d1', SellarDis1withDerivatives(), promotes=['x', 'z', 'y1', 'y2'])
        sub.add_subsystem('d2', SellarDis2withDerivatives(), promotes=['z', 'y1', 'y2'])

        model.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                z=np.array([0.0, 0.0]), x=0.0),
                            promotes=['obj', 'x', 'z', 'y1', 'y2'])

        model.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'), promotes=['con1', 'y1'])
        model.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'), promotes=['con2', 'y2'])

        sub.nonlinear_solver = om.BroydenSolver()
        sub.linear_solver = om.DirectSolver()
        model.linear_solver = om.DirectSolver()

        prob.model.add_design_var('x', lower=-100, upper=100)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.model.add_objective('obj')
        prob.model.add_constraint('con1', upper=0.0)
        prob.model.add_constraint('con2', upper=0.0)

        prob.setup(check=False, force_alloc_complex=True)
        prob.set_solver_print(level=0)

        prob.run_model()

        sub.nonlinear_solver.options['compute_jacobian'] = True

        totals = prob.check_totals(method='cs', out_stream=None)

        for key, val in totals.items():
            assert_near_equal(val['rel error'][0], 0.0, 1e-6)

    def test_complex_step(self):
        prob = om.Problem()
        model = prob.model
        sub = model.add_subsystem('sub', om.ParallelGroup(), promotes=['*'])

        model.add_subsystem('px', om.IndepVarComp('x', 1.0), promotes=['x'])
        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])), promotes=['z'])

        sub.add_subsystem('d1', SellarDis1withDerivatives(), promotes=['x', 'z', 'y1', 'y2'])
        sub.add_subsystem('d2', SellarDis2withDerivatives(), promotes=['z', 'y1', 'y2'])

        model.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                z=np.array([0.0, 0.0]), x=0.0),
                            promotes=['obj', 'x', 'z', 'y1', 'y2'])

        model.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'),
                            promotes=['con1', 'y1'])
        model.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'),
                            promotes=['con2', 'y2'])

        sub.nonlinear_solver = om.BroydenSolver()
        sub.linear_solver = om.DirectSolver()
        model.linear_solver = om.DirectSolver()

        prob.model.add_design_var('x', lower=-100, upper=100)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.model.add_objective('obj')
        prob.model.add_constraint('con1', upper=0.0)
        prob.model.add_constraint('con2', upper=0.0)

        prob.setup(check=False, force_alloc_complex=True)
        prob.set_solver_print(level=0)

        prob.run_model()

        totals = prob.check_totals(method='cs', out_stream=None)

        for key, val in totals.items():
            assert_near_equal(val['rel error'][0], 0.0, 1e-7)


# Commented the following test out until we fix the broyden check
# @unittest.skipUnless(MPI and PETScVector, "only run with MPI and PETSc.")
# class TestBryodenMPI(unittest.TestCase):

#     N_PROCS = 2

#     def test_distributed_comp(self):
#         class Y1Comp(om.ExplicitComponent):
#             def __init__(self, arr_size=11, **kwargs):
#                 super().__init__(**kwargs)
#                 self.arr_size = arr_size
#                 self.options['distributed'] = True

#             def setup(self):
#                 comm = self.comm
#                 rank = comm.rank

#                 sizes, _ = evenly_distrib_idxs(comm.size, self.arr_size)

#                 self.add_input('y2', np.ones(sizes[rank]))
#                 self.add_output('y1', np.ones(sizes[rank]))

#                 self.declare_partials(of='y1', wrt='y2', method='cs')

#             def compute(self, inputs, outputs):
#                 if self.comm.rank == 0:
#                     outputs['y1'] = 28. - .2 * inputs['y2']
#                 else:
#                     outputs['y1'] = 18. - .2 * inputs['y2']

#         class Y2Comp(om.ExplicitComponent):
#             def __init__(self, arr_size=11, **kwargs):
#                 super().__init__(**kwargs)
#                 self.arr_size = arr_size
#                 self.options['distributed'] = True

#             def setup(self):
#                 comm = self.comm
#                 rank = comm.rank

#                 sizes, _ = evenly_distrib_idxs(comm.size, self.arr_size)

#                 self.add_input('y1', np.ones(sizes[rank]))
#                 self.add_output('y2', np.ones(sizes[rank]))

#                 self.declare_partials(of='y2', wrt='y1', method='cs')

#             def compute(self, inputs, outputs):
#                 if self.comm.rank == 0:
#                     outputs['y2'] = inputs['y1'] ** .5 + 7.
#                 else:
#                     outputs['y2'] = inputs['y1'] ** .5 - 3.

#         prob = om.Problem()
#         model = prob.model
#         sub = model.add_subsystem('sub', om.Group(), promotes=['*'])

#         sub.add_subsystem('d1', Y1Comp(arr_size=2), promotes=['y1', 'y2'])
#         sub.add_subsystem('d2', Y2Comp(arr_size=2), promotes=['y1', 'y2'])

#         sub.nonlinear_solver = om.BroydenSolver()
#         sub.linear_solver = om.LinearBlockGS()
#         model.linear_solver = om.LinearBlockGS()

#         prob.setup(check=False, force_alloc_complex=True)

#         with self.assertRaises(Exception) as cm:
#             prob.run_model()

#         msg = "BroydenSolver linear solver in Group (sub) cannot be used in or above a ParallelGroup or a " + \
#             "distributed component."
#         self.assertEqual(str(cm.exception), msg)


class TestBryodenFeature(unittest.TestCase):

    def test_circuit(self):

        p = om.Problem()
        model = p.model

        model.add_subsystem('circuit', Circuit(), promotes_inputs=[('Vg', 'V'), ('I_in', 'I')])
        model.set_input_defaults('V', 0., units='V')
        model.set_input_defaults('I', 0.1, units='A')

        p.setup()

        # Replace existing solver with om.BroydenSolver
        model.circuit.nonlinear_solver = om.BroydenSolver()
        model.circuit.nonlinear_solver.options['maxiter'] = 20

        # Specify states for Broyden to solve
        model.circuit.nonlinear_solver.options['state_vars'] = ['n1.V', 'n2.V']

        model.nonlinear_solver.linear_solver = om.LinearBlockGS()

        # set some initial guesses
        p.set_val('circuit.n1.V', 10.)
        p.set_val('circuit.n2.V', 1.)

        p.set_solver_print(level=2)
        p.run_model()

        assert_near_equal(p.get_val('circuit.n1.V'), 9.90804735, 1e-5)
        assert_near_equal(p.get_val('circuit.n2.V'), 0.71278226, 1e-5)

        # sanity check: should sum to .1 Amps
        assert_near_equal(p.get_val('circuit.R1.I') + p.get_val('circuit.D1.I'), .1, 1e-6)

    def test_circuit_options(self):

        p = om.Problem()
        model = p.model

        model.add_subsystem('circuit', Circuit(), promotes_inputs=[('Vg', 'V'), ('I_in', 'I')])
        model.set_input_defaults('V', 0., units='V')
        model.set_input_defaults('I', 0.1, units='A')

        p.setup()

        # Replace existing solver with BroydenSolver
        model.circuit.nonlinear_solver = om.BroydenSolver()
        model.circuit.nonlinear_solver.options['maxiter'] = 20
        model.circuit.nonlinear_solver.options['converge_limit'] = 0.1
        model.circuit.nonlinear_solver.options['max_converge_failures'] = 1

        # Specify states for Broyden to solve
        model.circuit.nonlinear_solver.options['state_vars'] = ['n1.V', 'n2.V']

        # set some initial guesses
        p.set_val('circuit.n1.V', 10.)
        p.set_val('circuit.n2.V', 1.)

        p.set_solver_print(level=2)
        p.run_model()

        assert_near_equal(p.get_val('circuit.n1.V'), 9.90804735, 1e-5)
        assert_near_equal(p.get_val('circuit.n2.V'), 0.71278226, 1e-5)

        # sanity check: should sum to .1 Amps
        assert_near_equal(p.get_val('circuit.R1.I') + p.get_val('circuit.D1.I'), .1, 1e-6)

    def test_circuit_full(self):

        p = om.Problem()
        model = p.model

        model.add_subsystem('circuit', Circuit(), promotes_inputs=[('Vg', 'V'), ('I_in', 'I')])
        model.set_input_defaults('V', 0., units='V')
        model.set_input_defaults('I', 0.1, units='A')

        p.setup()

        # Replace existing solver with BroydenSolver
        model.circuit.nonlinear_solver = om.BroydenSolver()
        model.circuit.nonlinear_solver.options['maxiter'] = 20
        model.circuit.nonlinear_solver.linear_solver = om.DirectSolver()

        # set some initial guesses
        p.set_val('circuit.n1.V', 10.)
        p.set_val('circuit.n2.V', 1.)

        p.set_solver_print(level=2)
        p.run_model()

        assert_near_equal(p.get_val('circuit.n1.V'), 9.90804735, 1e-5)
        assert_near_equal(p.get_val('circuit.n2.V'), 0.71278226, 1e-5)

        # sanity check: should sum to .1 Amps
        assert_near_equal(p.get_val('circuit.R1.I') + p.get_val('circuit.D1.I'), .1, 1e-6)


if __name__ == "__main__":
    unittest.main()
